# -*- coding: utf-8 -*-
"""
biosppy.signals.tools
---------------------

This module provides various signal analysis methods in the time and
frequency domains.

:copyright: (c) 2015-2018 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function
from six.moves import range
import six

# 3rd party
import sys
import numpy as np
import scipy.signal as ss
from scipy import interpolate, optimize
from scipy.stats import stats

# local
from .. import utils


def _norm_freq(frequency=None, sampling_rate=1000.):
    """Normalize frequency to Nyquist Frequency (Fs/2).

    Parameters
    ----------
    frequency : int, float, list, array
        Frequencies to normalize.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).

    Returns
    -------
    wn : float, array
        Normalized frequencies.

    """

    # check inputs
    if frequency is None:
        raise TypeError("Please specify a frequency to normalize.")

    # convert inputs to correct representation
    try:
        frequency = float(frequency)
    except TypeError:
        # maybe frequency is a list or array
        frequency = np.array(frequency, dtype='float')

    Fs = float(sampling_rate)

    wn = 2. * frequency / Fs

    return wn


def _filter_init(b, a, alpha=1.):
    """Get an initial filter state that corresponds to the steady-state
    of the step response.

    Parameters
    ----------
    b : array
        Numerator coefficients.
    a : array
        Denominator coefficients.
    alpha : float, optional
        Scaling factor.

    Returns
    -------
    zi : array
        Initial filter state.

    """

    zi = alpha * ss.lfilter_zi(b, a)

    return zi


def _filter_signal(b, a, signal, zi=None, check_phase=True, **kwargs):
    """Filter a signal with given coefficients.

    Parameters
    ----------
    b : array
        Numerator coefficients.
    a : array
        Denominator coefficients.
    signal : array
        Signal to filter.
    zi : array, optional
        Initial filter state.
    check_phase : bool, optional
        If True, use the forward-backward technique.
    ``**kwargs`` : dict, optional
        Additional keyword arguments are passed to the underlying filtering
        function.

    Returns
    -------
    filtered : array
        Filtered signal.
    zf : array
        Final filter state.

    Notes
    -----
    * If check_phase is True, zi cannot be set.

    """

    # check inputs
    if check_phase and zi is not None:
        raise ValueError(
            "Incompatible arguments: initial filter state cannot be set when \
            check_phase is True.")

    if zi is None:
        zf = None
        if check_phase:
            filtered = ss.filtfilt(b, a, signal, **kwargs)
        else:
            filtered = ss.lfilter(b, a, signal, **kwargs)
    else:
        filtered, zf = ss.lfilter(b, a, signal, zi=zi, **kwargs)

    return filtered, zf


def _filter_resp(b, a, sampling_rate=1000., nfreqs=4096):
    """Compute the filter frequency response.

    Parameters
    ----------
    b : array
        Numerator coefficients.
    a : array
        Denominator coefficients.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    nfreqs : int, optional
        Number of frequency points to compute.

    Returns
    -------
    freqs : array
        Array of frequencies (Hz) at which the response was computed.
    resp : array
        Frequency response.

    """

    w, resp = ss.freqz(b, a, nfreqs)

    # convert frequencies
    freqs = w * sampling_rate / (2. * np.pi)

    return freqs, resp


def _get_window(kernel, size, **kwargs):
    """Return a window with the specified parameters.

    Parameters
    ----------
    kernel : str
        Type of window to create.
    size : int
        Size of the window.
    ``**kwargs`` : dict, optional
        Additional keyword arguments are passed to the underlying
        scipy.signal.windows function.

    Returns
    -------
    window : array
        Created window.

    """

    # mimics scipy.signal.get_window
    if kernel in ['blackman', 'black', 'blk']:
        winfunc = ss.blackman
    elif kernel in ['triangle', 'triang', 'tri']:
        winfunc = ss.triang
    elif kernel in ['hamming', 'hamm', 'ham']:
        winfunc = ss.hamming
    elif kernel in ['bartlett', 'bart', 'brt']:
        winfunc = ss.bartlett
    elif kernel in ['hanning', 'hann', 'han']:
        winfunc = ss.hann
    elif kernel in ['blackmanharris', 'blackharr', 'bkh']:
        winfunc = ss.blackmanharris
    elif kernel in ['parzen', 'parz', 'par']:
        winfunc = ss.parzen
    elif kernel in ['bohman', 'bman', 'bmn']:
        winfunc = ss.bohman
    elif kernel in ['nuttall', 'nutl', 'nut']:
        winfunc = ss.nuttall
    elif kernel in ['barthann', 'brthan', 'bth']:
        winfunc = ss.barthann
    elif kernel in ['flattop', 'flat', 'flt']:
        winfunc = ss.flattop
    elif kernel in ['kaiser', 'ksr']:
        winfunc = ss.kaiser
    elif kernel in ['gaussian', 'gauss', 'gss']:
        winfunc = ss.gaussian
    elif kernel in ['general gaussian', 'general_gaussian', 'general gauss',
                    'general_gauss', 'ggs']:
        winfunc = ss.general_gaussian
    elif kernel in ['boxcar', 'box', 'ones', 'rect', 'rectangular']:
        winfunc = ss.boxcar
    elif kernel in ['slepian', 'slep', 'optimal', 'dpss', 'dss']:
        winfunc = ss.slepian
    elif kernel in ['cosine', 'halfcosine']:
        winfunc = ss.cosine
    elif kernel in ['chebwin', 'cheb']:
        winfunc = ss.chebwin
    else:
        raise ValueError("Unknown window type.")

    try:
        window = winfunc(size, **kwargs)
    except TypeError as e:
        raise TypeError("Invalid window arguments: %s." % e)

    return window


def get_filter(ftype='FIR',
               band='lowpass',
               order=None,
               frequency=None,
               sampling_rate=1000., **kwargs):
    """Compute digital (FIR or IIR) filter coefficients with the given
    parameters.

    Parameters
    ----------
    ftype : str
        Filter type:
            * Finite Impulse Response filter ('FIR');
            * Butterworth filter ('butter');
            * Chebyshev filters ('cheby1', 'cheby2');
            * Elliptic filter ('ellip');
            * Bessel filter ('bessel').
    band : str
        Band type:
            * Low-pass filter ('lowpass');
            * High-pass filter ('highpass');
            * Band-pass filter ('bandpass');
            * Band-stop filter ('bandstop').
    order : int
        Order of the filter.
    frequency : int, float, list, array
        Cutoff frequencies; format depends on type of band:
            * 'lowpass' or 'highpass': single frequency;
            * 'bandpass' or 'bandstop': pair of frequencies.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    ``**kwargs`` : dict, optional
        Additional keyword arguments are passed to the underlying
        scipy.signal function.

    Returns
    -------
    b : array
        Numerator coefficients.
    a : array
        Denominator coefficients.

    See Also:
        scipy.signal

    """

    # check inputs
    if order is None:
        raise TypeError("Please specify the filter order.")
    if frequency is None:
        raise TypeError("Please specify the cutoff frequency.")
    if band not in ['lowpass', 'highpass', 'bandpass', 'bandstop']:
        raise ValueError(
            "Unknown filter type '%r'; choose 'lowpass', 'highpass', \
            'bandpass', or 'bandstop'."
            % band)

    # convert frequencies
    frequency = _norm_freq(frequency, sampling_rate)

    # get coeffs
    b, a = [], []
    if ftype == 'FIR':
        # FIR filter
        if order % 2 == 0:
            order += 1
        a = np.array([1])
        if band in ['lowpass', 'bandstop']:
            b = ss.firwin(numtaps=order,
                          cutoff=frequency,
                          pass_zero=True, **kwargs)
        elif band in ['highpass', 'bandpass']:
            b = ss.firwin(numtaps=order,
                          cutoff=frequency,
                          pass_zero=False, **kwargs)
    elif ftype == 'butter':
        # Butterworth filter
        b, a = ss.butter(N=order,
                         Wn=frequency,
                         btype=band,
                         analog=False,
                         output='ba', **kwargs)
    elif ftype == 'cheby1':
        # Chebyshev type I filter
        b, a = ss.cheby1(N=order,
                         Wn=frequency,
                         btype=band,
                         analog=False,
                         output='ba', **kwargs)
    elif ftype == 'cheby2':
        # chebyshev type II filter
        b, a = ss.cheby2(N=order,
                         Wn=frequency,
                         btype=band,
                         analog=False,
                         output='ba', **kwargs)
    elif ftype == 'ellip':
        # Elliptic filter
        b, a = ss.ellip(N=order,
                        Wn=frequency,
                        btype=band,
                        analog=False,
                        output='ba', **kwargs)
    elif ftype == 'bessel':
        # Bessel filter
        b, a = ss.bessel(N=order,
                         Wn=frequency,
                         btype=band,
                         analog=False,
                         output='ba', **kwargs)

    return utils.ReturnTuple((b, a), ('b', 'a'))


def filter_signal(signal=None,
                  ftype='FIR',
                  band='lowpass',
                  order=None,
                  frequency=None,
                  sampling_rate=1000., **kwargs):
    """Filter a signal according to the given parameters.

    Parameters
    ----------
    signal : array
        Signal to filter.
    ftype : str
        Filter type:
            * Finite Impulse Response filter ('FIR');
            * Butterworth filter ('butter');
            * Chebyshev filters ('cheby1', 'cheby2');
            * Elliptic filter ('ellip');
            * Bessel filter ('bessel').
    band : str
        Band type:
            * Low-pass filter ('lowpass');
            * High-pass filter ('highpass');
            * Band-pass filter ('bandpass');
            * Band-stop filter ('bandstop').
    order : int
        Order of the filter.
    frequency : int, float, list, array
        Cutoff frequencies; format depends on type of band:
            * 'lowpass' or 'bandpass': single frequency;
            * 'bandpass' or 'bandstop': pair of frequencies.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    ``**kwargs`` : dict, optional
        Additional keyword arguments are passed to the underlying
        scipy.signal function.

    Returns
    -------
    signal : array
        Filtered signal.
    sampling_rate : float
        Sampling frequency (Hz).
    params : dict
        Filter parameters.

    Notes
    -----
    * Uses a forward-backward filter implementation. Therefore, the combined
      filter has linear phase.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify a signal to filter.")

    # get filter
    b, a = get_filter(ftype=ftype,
                      order=order,
                      frequency=frequency,
                      sampling_rate=sampling_rate,
                      band=band, **kwargs)

    # filter
    filtered, _ = _filter_signal(b, a, signal, check_phase=True)

    # output
    params = {
        'ftype': ftype,
        'order': order,
        'frequency': frequency,
        'band': band,
    }
    params.update(kwargs)

    args = (filtered, sampling_rate, params)
    names = ('signal', 'sampling_rate', 'params')

    return utils.ReturnTuple(args, names)


class OnlineFilter(object):
    """Online filtering.

    Parameters
    ----------
    b : array
        Numerator coefficients.
    a : array
        Denominator coefficients.

    """

    def __init__(self, b=None, a=None):
        # check inputs
        if b is None:
            raise TypeError('Please specify the numerator coefficients.')

        if a is None:
            raise TypeError('Please specify the denominator coefficients.')

        # self things
        self.b = b
        self.a = a

        # reset
        self.reset()

    def reset(self):
        """Reset the filter state."""

        self.zi = None

    def filter(self, signal=None):
        """Filter a signal segment.

        Parameters
        ----------
        signal : array
            Signal segment to filter.

        Returns
        -------
        filtered : array
            Filtered signal segment.

        """

        # check input
        if signal is None:
            raise TypeError('Please specify the input signal.')

        if self.zi is None:
            self.zi = signal[0] * ss.lfilter_zi(self.b, self.a)

        filtered, self.zi = ss.lfilter(self.b, self.a, signal, zi=self.zi)

        return utils.ReturnTuple((filtered, ), ('filtered', ))


def smoother(signal=None, kernel='boxzen', size=10, mirror=True, **kwargs):
    """Smooth a signal using an N-point moving average [MAvg]_ filter.

    This implementation uses the convolution of a filter kernel with the input
    signal to compute the smoothed signal [Smit97]_.

    Availabel kernels: median, boxzen, boxcar, triang, blackman, hamming, hann,
    bartlett, flattop, parzen, bohman, blackmanharris, nuttall, barthann,
    kaiser (needs beta), gaussian (needs std), general_gaussian (needs power,
    width), slepian (needs width), chebwin (needs attenuation).

    Parameters
    ----------
    signal : array
        Signal to smooth.
    kernel : str, array, optional
        Type of kernel to use; if array, use directly as the kernel.
    size : int, optional
        Size of the kernel; ignored if kernel is an array.
    mirror : bool, optional
        If True, signal edges are extended to avoid boundary effects.
    ``**kwargs`` : dict, optional
        Additional keyword arguments are passed to the underlying
        scipy.signal.windows function.

    Returns
    -------
    signal : array
        Smoothed signal.
    params : dict
        Smoother parameters.

    Notes
    -----
    * When the kernel is 'median', mirror is ignored.

    References
    ----------
    .. [MAvg] Wikipedia, "Moving Average",
       http://en.wikipedia.org/wiki/Moving_average
    .. [Smit97] S. W. Smith, "Moving Average Filters - Implementation by
       Convolution", http://www.dspguide.com/ch15/1.htm, 1997

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify a signal to smooth.")

    length = len(signal)

    if isinstance(kernel, six.string_types):
        # check length
        if size > length:
            size = length - 1

        if size < 1:
            size = 1

        if kernel == 'boxzen':
            # hybrid method
            # 1st pass - boxcar kernel
            aux, _ = smoother(signal,
                              kernel='boxcar',
                              size=size,
                              mirror=mirror)

            # 2nd pass - parzen kernel
            smoothed, _ = smoother(aux,
                                   kernel='parzen',
                                   size=size,
                                   mirror=mirror)

            params = {'kernel': kernel, 'size': size, 'mirror': mirror}

            args = (smoothed, params)
            names = ('signal', 'params')

            return utils.ReturnTuple(args, names)

        elif kernel == 'median':
            # median filter
            if size % 2 == 0:
                raise ValueError(
                    "When the kernel is 'median', size must be odd.")

            smoothed = ss.medfilt(signal, kernel_size=size)

            params = {'kernel': kernel, 'size': size, 'mirror': mirror}

            args = (smoothed, params)
            names = ('signal', 'params')

            return utils.ReturnTuple(args, names)

        else:
            win = _get_window(kernel, size, **kwargs)

    elif isinstance(kernel, np.ndarray):
        win = kernel
        size = len(win)

        # check length
        if size > length:
            raise ValueError("Kernel size is bigger than signal length.")

        if size < 1:
            raise ValueError("Kernel size is smaller than 1.")

    else:
        raise TypeError("Unknown kernel type.")

    # convolve
    w = win / win.sum()
    if mirror:
        aux = np.concatenate(
            (signal[0] * np.ones(size), signal, signal[-1] * np.ones(size)))
        smoothed = np.convolve(w, aux, mode='same')
        smoothed = smoothed[size:-size]
    else:
        smoothed = np.convolve(w, signal, mode='same')

    # output
    params = {'kernel': kernel, 'size': size, 'mirror': mirror}
    params.update(kwargs)

    args = (smoothed, params)
    names = ('signal', 'params')

    return utils.ReturnTuple(args, names)


def analytic_signal(signal=None, N=None):
    """Compute analytic signal, using the Hilbert Transform.

    Parameters
    ----------
    signal : array
        Input signal.
    N : int, optional
        Number of Fourier components; default is `len(signal)`.

    Returns
    -------
    amplitude : array
        Amplitude envelope of the analytic signal.
    phase : array
        Instantaneous phase component of the analystic signal.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # hilbert transform
    asig = ss.hilbert(signal, N=N)

    # amplitude envelope
    amp = np.absolute(asig)

    # instantaneous
    phase = np.angle(asig)

    return utils.ReturnTuple((amp, phase), ('amplitude', 'phase'))


def phase_locking(signal1=None, signal2=None, N=None):
    """Compute the Phase-Locking Factor (PLF) between two signals.

    Parameters
    ----------
    signal1 : array
        First input signal.
    signal2 : array
        Second input signal.
    N : int, optional
        Number of Fourier components.

    Returns
    -------
    plf : float
        The PLF between the two signals.

    """

    # check inputs
    if signal1 is None:
        raise TypeError("Please specify the first input signal.")

    if signal2 is None:
        raise TypeError("Please specify the second input signal.")

    if len(signal1) != len(signal2):
        raise ValueError("The input signals must have the same length.")

    # compute analytic signal
    asig1 = ss.hilbert(signal1, N=N)
    phase1 = np.angle(asig1)

    asig2 = ss.hilbert(signal2, N=N)
    phase2 = np.angle(asig2)

    # compute PLF
    plf = np.absolute(np.mean(np.exp(1j * (phase1 - phase2))))

    return utils.ReturnTuple((plf,), ('plf',))


def power_spectrum(signal=None,
                   sampling_rate=1000.,
                   pad=None,
                   pow2=False,
                   decibel=True):
    """Compute the power spectrum of a signal (one-sided).

    Parameters
    ----------
    signal : array
        Input signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    pad : int, optional
        Padding for the Fourier Transform (number of zeros added);
        defaults to no padding..
    pow2 : bool, optional
        If True, rounds the number of points `N = len(signal) + pad` to the
        nearest power of 2 greater than N.
    decibel : bool, optional
        If True, returns the power in decibels.

    Returns
    -------
    freqs : array
        Array of frequencies (Hz) at which the power was computed.
    power : array
        Power spectrum.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    npoints = len(signal)

    if pad is not None:
        if pad >= 0:
            npoints += pad
        else:
            raise ValueError("Padding must be a positive integer.")

    # power of 2
    if pow2:
        npoints = 2 ** (np.ceil(np.log2(npoints)))

    Nyq = float(sampling_rate) / 2
    hpoints = npoints // 2

    freqs = np.linspace(0, Nyq, hpoints)
    power = np.abs(np.fft.fft(signal, npoints)) / npoints

    # one-sided
    power = power[:hpoints]
    power[1:] *= 2
    power = np.power(power, 2)

    if decibel:
        power = 10. * np.log10(power)

    return utils.ReturnTuple((freqs, power), ('freqs', 'power'))


def welch_spectrum(signal=None,
                   sampling_rate=1000.,
                   size=None,
                   overlap=None,
                   window='hanning',
                   window_kwargs=None,
                   pad=None,
                   decibel=True):
    """Compute the power spectrum of a signal using Welch's method (one-sided).

    Parameters
    ----------
    signal : array
        Input signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    size : int, optional
        Number of points in each Welch segment;
        defaults to the equivalent of 1 second;
        ignored when 'window' is an array.
    overlap : int, optional
        Number of points to overlap between segments; defaults to `size / 2`.
    window : str, array, optional
        Type of window to use.
    window_kwargs : dict, optional
        Additional keyword arguments to pass on window creation; ignored if
        'window' is an array.
    pad : int, optional
        Padding for the Fourier Transform (number of zeros added);
        defaults to no padding.
    decibel : bool, optional
        If True, returns the power in decibels.

    Returns
    -------
    freqs : array
        Array of frequencies (Hz) at which the power was computed.
    power : array
        Power spectrum.

    Notes
    -----
    * Detrends each Welch segment by removing the mean.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    length = len(signal)
    sampling_rate = float(sampling_rate)

    if size is None:
        size = int(sampling_rate)

    if window_kwargs is None:
        window_kwargs = {}

    if isinstance(window, six.string_types):
        win = _get_window(window, size, **window_kwargs)
    elif isinstance(window, np.ndarray):
        win = window
        size = len(win)

    if size > length:
        raise ValueError('Segment size must be smaller than signal length.')

    if overlap is None:
        overlap = size // 2
    elif overlap > size:
        raise ValueError('Overlap must be smaller than segment size.')

    nfft = size
    if pad is not None:
        if pad >= 0:
            nfft += pad
        else:
            raise ValueError("Padding must be a positive integer.")

    freqs, power = ss.welch(
        signal,
        fs=sampling_rate,
        window=win,
        nperseg=size,
        noverlap=overlap,
        nfft=nfft,
        detrend='constant',
        return_onesided=True,
        scaling='spectrum',
    )

    # compensate one-sided
    power *= 2

    if decibel:
        power = 10. * np.log10(power)

    return utils.ReturnTuple((freqs, power), ('freqs', 'power'))


def band_power(freqs=None, power=None, frequency=None, decibel=True):
    """Compute the avearge power in a frequency band.

    Parameters
    ----------
    freqs : array
        Array of frequencies (Hz) at which the power was computed.
    power : array
        Input power spectrum.
    frequency : list, array
        Pair of frequencies defining the band.
    decibel : bool, optional
        If True, input power is in decibels.

    Returns
    -------
    avg_power : float
        The average power in the band.

    """

    # check inputs
    if freqs is None:
        raise TypeError("Please specify the 'freqs' array.")

    if power is None:
        raise TypeError("Please specify the input power spectrum.")

    if len(freqs) != len(power):
        raise ValueError(
            "The input 'freqs' and 'power' arrays must have the same length.")

    if frequency is None:
        raise TypeError("Please specify the band frequencies.")

    try:
        f1, f2 = frequency
    except ValueError:
        raise ValueError("Input 'frequency' must be a pair of frequencies.")

    # make frequencies sane
    if f1 > f2:
        f1, f2 = f2, f1

    if f1 < freqs[0]:
        f1 = freqs[0]
    if f2 > freqs[-1]:
        f2 = freqs[-1]

    # average
    sel = np.nonzero(np.logical_and(f1 <= freqs, freqs <= f2))[0]

    if decibel:
        aux = 10 ** (power / 10.)
        avg = np.mean(aux[sel])
        avg = 10. * np.log10(avg)
    else:
        avg = np.mean(power[sel])

    return utils.ReturnTuple((avg,), ('avg_power',))


def signal_stats(signal=None):
    """Compute various metrics describing the signal.

    Parameters
    ----------
    signal : array
        Input signal.

    Returns
    -------
    mean : float
        Mean of the signal.
    median : float
        Median of the signal.
    min : float
        Minimum signal value.
    max : float
        Maximum signal value.
    max_amp : float
        Maximum absolute signal amplitude, in relation to the mean.
    var : float
        Signal variance (unbiased).
    std_dev : float
        Standard signal deviation (unbiased).
    abs_dev : float
        Mean absolute signal deviation around the median.
    kurtosis : float
        Signal kurtosis (unbiased).
    skew : float
        Signal skewness (unbiased).

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    # mean
    mean = np.mean(signal)

    # median
    median = np.median(signal)

    # min
    minVal = np.min(signal)

    # max
    maxVal = np.max(signal)

    # maximum amplitude
    maxAmp = np.abs(signal - mean).max()

    # variance
    sigma2 = signal.var(ddof=1)

    # standard deviation
    sigma = signal.std(ddof=1)

    # absolute deviation
    ad = np.mean(np.abs(signal - median))

    # kurtosis
    kurt = stats.kurtosis(signal, bias=False)

    # skweness
    skew = stats.skew(signal, bias=False)

    # output
    args = (mean, median, minVal, maxVal, maxAmp, sigma2, sigma, ad, kurt, skew)
    names = ('mean', 'median', 'min', 'max', 'max_amp', 'var', 'std_dev',
             'abs_dev', 'kurtosis', 'skewness')

    return utils.ReturnTuple(args, names)


def normalize(signal=None, ddof=1):
    """Normalize a signal to zero mean and unitary standard deviation.

    Parameters
    ----------
    signal : array
        Input signal.
    ddof : int, optional
        Delta degrees of freedom for standard deviation computation;
        the divisor is `N - ddof`, where `N` is the number of elements;
        default is one.

    Returns
    -------
    signal : array
        Normalized signal.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    normalized = signal - signal.mean()
    normalized /= normalized.std(ddof=ddof)

    return utils.ReturnTuple((normalized,), ('signal',))


def zero_cross(signal=None, detrend=False):
    """Locate the indices where the signal crosses zero.

    Parameters
    ----------
    signal : array
        Input signal.
    detrend : bool, optional
        If True, remove signal mean before computation.

    Returns
    -------
    zeros : array
        Indices of zero crossings.

    Notes
    -----
    * When the signal crosses zero between samples, the first index
      is returned.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    if detrend:
        signal = signal - np.mean(signal)

    # zeros
    df = np.diff(np.sign(signal))
    zeros = np.nonzero(np.abs(df) > 0)[0]

    return utils.ReturnTuple((zeros,), ('zeros',))


def find_extrema(signal=None, mode='both'):
    """Locate local extrema points in a signal.

    Based on Fermat's Theorem [Ferm]_.

    Parameters
    ----------
    signal : array
        Input signal.
    mode : str, optional
        Whether to find maxima ('max'), minima ('min'), or both ('both').

    Returns
    -------
    extrema : array
        Indices of the extrama points.
    values : array
        Signal values at the extrema points.

    References
    ----------
    .. [Ferm] Wikipedia, "Fermat's theorem (stationary points)",
       https://en.wikipedia.org/wiki/Fermat%27s_theorem_(stationary_points)

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    if mode not in ['max', 'min', 'both']:
        raise ValueError("Unknwon mode %r." % mode)

    aux = np.diff(np.sign(np.diff(signal)))

    if mode == 'both':
        aux = np.abs(aux)
        extrema = np.nonzero(aux > 0)[0] + 1
    elif mode == 'max':
        extrema = np.nonzero(aux < 0)[0] + 1
    elif mode == 'min':
        extrema = np.nonzero(aux > 0)[0] + 1

    values = signal[extrema]

    return utils.ReturnTuple((extrema, values), ('extrema', 'values'))


def windower(signal=None,
             size=None,
             step=None,
             fcn=None,
             fcn_kwargs=None,
             kernel='boxcar',
             kernel_kwargs=None):
    """Apply a function to a signal in sequential windows, with optional overlap.

    Availabel window kernels: boxcar, triang, blackman, hamming, hann,
    bartlett, flattop, parzen, bohman, blackmanharris, nuttall, barthann,
    kaiser (needs beta), gaussian (needs std), general_gaussian (needs power,
    width), slepian (needs width), chebwin (needs attenuation).

    Parameters
    ----------
    signal : array
        Input signal.
    size : int
        Size of the signal window.
    step : int, optional
        Size of window shift; if None, there is no overlap.
    fcn : callable
        Function to apply to each window.
    fcn_kwargs : dict, optional
        Additional keyword arguments to pass to 'fcn'.
    kernel : str, array, optional
        Type of kernel to use; if array, use directly as the kernel.
    kernel_kwargs : dict, optional
        Additional keyword arguments to pass on window creation; ignored if
        'kernel' is an array.

    Returns
    -------
    index : array
        Indices characterizing window locations (start of the window).
    values : array
        Concatenated output of calling 'fcn' on each window.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    if fcn is None:
        raise TypeError("Please specify a function to apply to each window.")

    if fcn_kwargs is None:
        fcn_kwargs = {}

    if kernel_kwargs is None:
        kernel_kwargs = {}

    length = len(signal)

    if isinstance(kernel, six.string_types):
        # check size
        if size > length:
            raise ValueError("Window size must be smaller than signal length.")

        win = _get_window(kernel, size, **kernel_kwargs)
    elif isinstance(kernel, np.ndarray):
        win = kernel
        size = len(win)

        # check size
        if size > length:
            raise ValueError("Window size must be smaller than signal length.")

    if step is None:
        step = size

    if step <= 0:
        raise ValueError("Step size must be at least 1.")

    # number of windows
    nb = 1 + (length - size) // step

    # check signal dimensionality
    if np.ndim(signal) == 2:
        # time along 1st dim, tile window
        nch = np.shape(signal)[1]
        win = np.tile(np.reshape(win, (size, 1)), nch)

    index = []
    values = []
    for i in range(nb):
        start = i * step
        stop = start + size
        index.append(start)

        aux = signal[start:stop] * win

        # apply function
        out = fcn(aux, **fcn_kwargs)
        values.append(out)

    # transform to numpy
    index = np.array(index, dtype='int')
    values = np.array(values)

    return utils.ReturnTuple((index, values), ('index', 'values'))


def synchronize(x=None, y=None, detrend=True):
    """Align two signals based on cross-correlation.

    Parameters
    ----------
    x : array
        First input signal.
    y : array
        Second input signal.
    detrend : bool, optional
        If True, remove signal means before computation.

    Returns
    -------
    delay : int
        Delay (number of samples) of 'x' in relation to 'y';
        if 'delay' < 0 , 'x' is ahead in relation to 'y';
        if 'delay' > 0 , 'x' is delayed in relation to 'y'.
    corr : float
        Value of maximum correlation.
    synch_x : array
        Biggest possible portion of 'x' in synchronization.
    synch_y : array
        Biggest possible portion of 'y' in synchronization.

    """

    # check inputs
    if x is None:
        raise TypeError("Please specify the first input signal.")

    if y is None:
        raise TypeError("Please specify the second input signal.")

    n1 = len(x)
    n2 = len(y)

    if detrend:
        x = x - np.mean(x)
        y = y - np.mean(y)

    # correlate
    corr = np.correlate(x, y, mode='full')
    d = np.arange(-n2 + 1, n1, dtype='int')
    ind = np.argmax(corr)

    delay = d[ind]
    maxCorr = corr[ind]

    # get synchronization overlap
    if delay < 0:
        c = min([n1, len(y[-delay:])])
        synch_x = x[:c]
        synch_y = y[-delay:-delay + c]
    elif delay > 0:
        c = min([n2, len(x[delay:])])
        synch_x = x[delay:delay + c]
        synch_y = y[:c]
    else:
        c = min([n1, n2])
        synch_x = x[:c]
        synch_y = y[:c]

    # output
    args = (delay, maxCorr, synch_x, synch_y)
    names = ('delay', 'corr', 'synch_x', 'synch_y')

    return utils.ReturnTuple(args, names)


def pearson_correlation(x=None, y=None):
    """Compute the Pearson Correlation Coefficient bertween two signals.

    The coefficient is given by:

    .. math::

        r_{xy} = \\frac{E[(X - \\mu_X) (Y - \\mu_Y)]}{\\sigma_X \\sigma_Y}

    Parameters
    ----------
    x : array
        First input signal.
    y : array
        Second input signal.

    Returns
    -------
    rxy : float
        Pearson correlation coefficient, ranging between -1 and +1.

    Raises
    ------
    ValueError
        If the input signals do not have the same length.

    """
    
    print("tools.pearson_correlation is deprecated, use stats.pearson_correlation instead", file=sys.stderr)

    # check inputs
    if x is None:
        raise TypeError("Please specify the first input signal.")

    if y is None:
        raise TypeError("Please specify the second input signal.")

    # ensure numpy
    x = np.array(x)
    y = np.array(y)

    n = len(x)

    if n != len(y):
        raise ValueError('Input signals must have the same length.')

    mx = np.mean(x)
    my = np.mean(y)

    Sxy = np.sum(x * y) - n*mx*my
    Sxx = np.sum(np.power(x, 2)) - n * mx**2
    Syy = np.sum(np.power(y, 2)) - n * my**2

    rxy = Sxy / (np.sqrt(Sxx) * np.sqrt(Syy))

    # avoid propagation of numerical errors
    if rxy > 1.0:
        rxy = 1.0
    elif rxy < -1.0:
        rxy = -1.0

    return utils.ReturnTuple((rxy, ), ('rxy', ))


def rms_error(x=None, y=None):
    """Compute the Root-Mean-Square Error between two signals.

    The error is given by:

    .. math::

        rmse = \\sqrt{E[(X - Y)^2]}

    Parameters
    ----------
    x : array
        First input signal.
    y : array
        Second input signal.

    Returns
    -------
    rmse : float
        Root-mean-square error.

    Raises
    ------
    ValueError
        If the input signals do not have the same length.

    """

    # check inputs
    if x is None:
        raise TypeError("Please specify the first input signal.")

    if y is None:
        raise TypeError("Please specify the second input signal.")

    # ensure numpy
    x = np.array(x)
    y = np.array(y)

    n = len(x)

    if n != len(y):
        raise ValueError('Input signals must have the same length.')

    rmse = np.sqrt(np.mean(np.power(x - y, 2)))

    return utils.ReturnTuple((rmse, ), ('rmse', ))


def get_heart_rate(beats=None, sampling_rate=1000., smooth=False, size=3):
    """Compute instantaneous heart rate from an array of beat indices.

    Parameters
    ----------
    beats : array
        Beat location indices.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    smooth : bool, optional
        If True, perform smoothing on the resulting heart rate.
    size : int, optional
        Size of smoothing window; ignored if `smooth` is False.

    Returns
    -------
    index : array
        Heart rate location indices.
    heart_rate : array
        Instantaneous heart rate (bpm).

    Notes
    -----
    * Assumes normal human heart rate to be between 40 and 200 bpm.

    """

    # check inputs
    if beats is None:
        raise TypeError("Please specify the input beat indices.")

    if len(beats) < 2:
        raise ValueError("Not enough beats to compute heart rate.")

    # compute heart rate
    ts = beats[1:]
    hr = sampling_rate * (60. / np.diff(beats))

    # physiological limits
    indx = np.nonzero(np.logical_and(hr >= 40, hr <= 200))
    ts = ts[indx]
    hr = hr[indx]

    # smooth with moving average
    if smooth and (len(hr) > 1):
        hr, _ = smoother(signal=hr, kernel='boxcar', size=size, mirror=True)

    return utils.ReturnTuple((ts, hr), ('index', 'heart_rate'))


def _pdiff(x, p1, p2):
    """Compute the squared difference between two interpolators, given the
    x-coordinates.

    Parameters
    ----------
    x : array
        Array of x-coordinates.
    p1 : object
        First interpolator.
    p2 : object
        Second interpolator.

    Returns
    -------
    diff : array
        Squared differences.

    """

    diff = (p1(x) - p2(x)) ** 2

    return diff


def find_intersection(x1=None,
                      y1=None,
                      x2=None,
                      y2=None,
                      alpha=1.5,
                      xtol=1e-6,
                      ytol=1e-6):
    """Find the intersection points between two lines using piecewise
    polynomial interpolation.

    Parameters
    ----------
    x1 : array
        Array of x-coordinates of the first line.
    y1 : array
        Array of y-coordinates of the first line.
    x2 : array
        Array of x-coordinates of the second line.
    y2 : array
        Array of y-coordinates of the second line.
    alpha : float, optional
        Resolution factor for the x-axis; fraction of total number of
        x-coordinates.
    xtol : float, optional
        Tolerance for the x-axis.
    ytol : float, optional
        Tolerance for the y-axis.

    Returns
    -------
    roots : array
        Array of x-coordinates of found intersection points.
    values : array
        Array of y-coordinates of found intersection points.

    Notes
    -----
    * If no intersection is found, returns the closest point.

    """

    # check inputs
    if x1 is None:
        raise TypeError("Please specify the x-coordinates of the first line.")
    if y1 is None:
        raise TypeError("Please specify the y-coordinates of the first line.")
    if x2 is None:
        raise TypeError("Please specify the x-coordinates of the second line.")
    if y2 is None:
        raise TypeError("Please specify the y-coordinates of the second line.")

    # ensure numpy
    x1 = np.array(x1)
    y1 = np.array(y1)
    x2 = np.array(x2)
    y2 = np.array(y2)

    if x1.shape != y1.shape:
        raise ValueError(
            "Input coordinates for the first line must have the same shape.")
    if x2.shape != y2.shape:
        raise ValueError(
            "Input coordinates for the second line must have the same shape.")

    # interpolate
    p1 = interpolate.BPoly.from_derivatives(x1, y1[:, np.newaxis])
    p2 = interpolate.BPoly.from_derivatives(x2, y2[:, np.newaxis])

    # combine x intervals
    x = np.r_[x1, x2]
    x_min = x.min()
    x_max = x.max()
    npoints = int(len(np.unique(x)) * alpha)
    x = np.linspace(x_min, x_max, npoints)

    # initial estimates
    pd = p1(x) - p2(x)
    zerocs, = zero_cross(pd)

    pd_abs = np.abs(pd)
    zeros = np.nonzero(pd_abs < ytol)[0]

    ind = np.unique(np.concatenate((zerocs, zeros)))
    xi = x[ind]

    # search for solutions
    roots = set()
    for v in xi:
        root, _, ier, _ = optimize.fsolve(
            _pdiff,
            v,
            args=(p1, p2),
            full_output=True,
            xtol=xtol,
        )
        if ier == 1 and x_min <= root <= x_max:
            roots.add(root[0])

    if len(roots) == 0:
        # no solution was found => give the best from the initial estimates
        aux = np.abs(pd)
        bux = aux.min() * np.ones(npoints, dtype='float')
        roots, _ = find_intersection(x, aux, x, bux,
                                     alpha=1.,
                                     xtol=xtol,
                                     ytol=ytol)

    # compute values
    roots = list(roots)
    roots.sort()
    roots = np.array(roots)
    values = np.mean(np.vstack((p1(roots), p2(roots))), axis=0)

    return utils.ReturnTuple((roots, values), ('roots', 'values'))


def finite_difference(signal=None, weights=None):
    """Apply the Finite Difference method to compute derivatives.
    
    Parameters
    ----------
    signal : array
        Signal to differentiate.
    weights : list, array
        Finite difference weight coefficients.
    
    Returns
    -------
    index : array
        Indices from `signal` for which the derivative was computed.
    derivative : array
        Computed derivative.
    
    Notes
    -----
    * The method assumes central differences weights.
    * The method accounts for the delay introduced by the algorithm.
    
    Raises
    ------
    ValueError
        If the number of weights is not odd.
    
    """
    
    # check inputs
    if signal is None:
        raise TypeError("Please specify a signal to differentiate.")
    
    if weights is None:
        raise TypeError("Please specify the weight coefficients.")
    
    N = len(weights)
    if N % 2 == 0:
        raise ValueError("Number of weights must be odd.")
    
    # diff
    weights = weights[::-1]
    derivative = ss.lfilter(weights, [1], signal)
    
    # trim delay
    D = N - 1
    D2 = D // 2
    
    index = np.arange(D2, len(signal) - D2, dtype='int')
    derivative = derivative[D:]
    
    return utils.ReturnTuple((index, derivative), ('index', 'derivative'))


def _init_dist_profile(m, n, signal):
    """Compute initial time series signal statistics for distance profile.
    
    Implements the algorithm described in [Mueen2014]_, using the notation
    from [Yeh2016_a]_.
    
    Parameters
    ----------
    m : int
        Sub-sequence length.
    n : int
        Target signal length.
    signal : array
        Target signal.
    
    Returns
    -------
    X : array
        Fourier Transform (FFT) of the signal.
    sigma : array
        Moving standard deviation in windows of length `m`.
    
    References
    ----------
    .. [Mueen2014] Abdullah Mueen, Hossein Hamooni, "Trilce Estrada: Time
       Series Join on Subsequence Correlation", ICDM 2014: 450-459
    .. [Yeh2016_a] Chin-Chia Michael Yeh, Yan Zhu, Liudmila Ulanova,
       Nurjahan Begum, Yifei Ding, Hoang Anh Dau, Diego Furtado Silva,
       Abdullah Mueen, Eamonn Keogh, "Matrix Profile I: All Pairs Similarity 
       Joins for Time Series: A Unifying View that Includes Motifs, Discords 
       and Shapelets", IEEE ICDM 2016
    
    """
    
    # compute signal stats
    csumx = np.zeros(n+1, dtype='float')
    csumx[1:] = np.cumsum(signal)
    sumx = csumx[m:] - csumx[:-m]
    
    csumx2 = np.zeros(n+1, dtype='float')
    csumx2[1:] = np.cumsum(np.power(signal, 2))
    sumx2 = csumx2[m:] - csumx2[:-m]
    
    meanx = sumx / m
    sigmax2 = (sumx2 / m) - np.power(meanx, 2)
    sigma = np.sqrt(sigmax2)
    
    # append zeros
    x = np.concatenate((signal, np.zeros(n, dtype='float')))
    
    # compute FFT
    X = np.fft.fft(x)
    
    return X, sigma


def _ditance_profile(m, n, query, X, sigma):
    """Compute the distance profile of a query sequence against a signal.
    
    Implements the algorithm described in [Mueen2014]_, using the notation
    from [Yeh2016]_.
    
    Parameters
    ----------
    m : int
        Query sub-sequence length.
    n : int
        Target time series length.
    query : array
        Query sub-sequence.
    X : array
        Target time series Fourier Transform (FFT).
    sigma : array
        Moving standard deviation in windows of length `m`.
    
    Returns
    -------
    dist : array
        Distance profile (squared).
    
    Notes
    -----
    * Computes distances on z-normalized data.
    
    References
    ----------
    .. [Mueen2014] Abdullah Mueen, Hossein Hamooni, "Trilce Estrada: Time
       Series Join on Subsequence Correlation", ICDM 2014: 450-459
    .. [Yeh2016] Chin-Chia Michael Yeh, Yan Zhu, Liudmila Ulanova,
       Nurjahan Begum, Yifei Ding, Hoang Anh Dau, Diego Furtado Silva,
       Abdullah Mueen, Eamonn Keogh, "Matrix Profile I: All Pairs Similarity 
       Joins for Time Series: A Unifying View that Includes Motifs, Discords 
       and Shapelets", IEEE ICDM 2016
    
    """
    
    # normalize query
    q = (query - np.mean(query)) / np.std(query)
    
    # reverse query and append zeros
    y = np.concatenate((q[::-1], np.zeros(2*n - m, dtype='float')))
    
    # compute dot products fast
    Y = np.fft.fft(y)
    Z = X * Y
    z = np.fft.ifft(Z)
    z = z[m-1:n]
    
    # compute distances (z-normalized squared euclidean distance)
    dist = 2 * m * (1 - z / (m * sigma))
    
    return dist


def distance_profile(query=None, signal=None, metric='euclidean'):
    """Compute the distance profile of a query sequence against a signal.
    
    Implements the algorithm described in [Mueen2014]_.
    
    Parameters
    ----------
    query : array
        Input query signal sequence.
    signal : array
        Input target time series signal.
    metric : str, optional
        The distance metric to use; one of 'euclidean' or 'pearson'; default
        is 'euclidean'.
    
    Returns
    -------
    dist : array
        Distance of the query sequence to every sub-sequnce in the signal.
    
    Notes
    -----
    * Computes distances on z-normalized data.
    
    References
    ----------
    .. [Mueen2014] Abdullah Mueen, Hossein Hamooni, "Trilce Estrada: Time
       Series Join on Subsequence Correlation", ICDM 2014: 450-459
    
    """
    
    # check inputs
    if query is None:
        raise TypeError("Please specify the input query sequence.")
    
    if signal is None:
        raise TypeError("Please specify the input time series signal.")
    
    if metric not in ['euclidean', 'pearson']:
        raise ValueError("Unknown distance metric.")
    
    # ensure numpy
    query = np.array(query)
    signal = np.array(signal)
    
    m = len(query)
    n = len(signal)
    if m > n/2:
        raise ValueError("Time series signal is too short relative to"
                         " query length.")
    
    # get initial signal stats
    X, sigma = _init_dist_profile(m, n, signal)
    
    # compute distance profile
    dist = _ditance_profile(m, n, query, X, sigma)
    
    if metric == 'pearson':
        dist = 1 - np.abs(dist) / (2 * m)
    elif metric == 'euclidean':
        dist = np.abs(np.sqrt(dist))
    
    return utils.ReturnTuple((dist, ), ('dist', ))


def signal_self_join(signal=None, size=None, index=None, limit=None):
    """Compute the matrix profile for a self-similarity join of a time series.
    
    Implements the algorithm described in [Yeh2016_b]_.
    
    Parameters
    ----------
    signal : array
        Input target time series signal.
    size : int
        Size of the query sub-sequences.
    index : list, array, optional
        Starting indices for query sub-sequences; the default is to search all
        sub-sequences.
    limit : int, optional
        Upper limit for the number of query sub-sequences; the default is to
        search all sub-sequences.
    
    Returns
    -------
    matrix_index : array
        Matric profile index.
    matrix_profile : array
        Computed matrix profile (distances).
    
    Notes
    -----
    * Computes euclidean distances on z-normalized data.
    
    References
    ----------
    .. [Yeh2016_b] Chin-Chia Michael Yeh, Yan Zhu, Liudmila Ulanova,
       Nurjahan Begum, Yifei Ding, Hoang Anh Dau, Diego Furtado Silva,
       Abdullah Mueen, Eamonn Keogh, "Matrix Profile I: All Pairs Similarity 
       Joins for Time Series: A Unifying View that Includes Motifs, Discords 
       and Shapelets", IEEE ICDM 2016
    
    """
    
    # check inputs
    if signal is None:
        raise TypeError("Please specify the input time series signal.")
    
    if size is None:
        raise TypeError("Please specify the sub-sequence size.")
    
    # ensure numpy
    signal = np.array(signal)
    
    n = len(signal)
    if size > n/2:
        raise ValueError("Time series signal is too short relative to desired"
                         " sub-sequence length.")
    
    if size < 4:
        raise ValueError("Sub-sequence length must be at least 4.")
    
    # matrix profile length
    nb = n - size + 1
    
    # get search index
    if index is None:
        index = np.random.permutation(np.arange(nb, dtype='int'))
    else:
        index = np.array(index)
        if not np.all(index < nb):
            raise ValueError("Provided `index` exceeds allowable sub-sequences.")
    
    # limit search
    if limit is not None:
        if limit < 1:
            raise ValueError("Search limit must be at least 1.")
        
        index = index[:limit]
    
    # exclusion zone (to avoid query self-matches)
    ezone = int(round(size / 4))
    
    # initialization
    matrix_profile = np.inf * np.ones(nb, dtype='float')
    matrix_index = np.zeros(nb, dtype='int')
    
    X, sigma = _init_dist_profile(size, n, signal)
    
    # compute matrix profile
    for idx in index:
        # compute distance profile
        query = signal[idx:idx+size]
        dist = _ditance_profile(size, n, query, X, sigma)
        dist = np.abs(np.sqrt(dist)) # to have euclidean distance
        
        # apply exlusion zone
        a = max([0, idx-ezone])
        b = min([nb, idx+ezone+1])
        dist[a:b] = np.inf
        
        # find nearest neighbors
        pos = dist < matrix_profile
        matrix_profile[pos] = dist[pos]
        matrix_index[pos] = idx
        
        # account for exlusion zone
        neighbor = np.argmin(dist)
        matrix_profile[idx] = dist[neighbor]
        matrix_index[idx] = neighbor
    
    # output
    args = (matrix_index, matrix_profile)
    names = ('matrix_index', 'matrix_profile')

    return utils.ReturnTuple(args, names)


def signal_cross_join(signal1=None,
                      signal2=None,
                      size=None,
                      index=None,
                      limit=None):
    """Compute the matrix profile for a similarity join of two time series.
    
    Computes the nearest sub-sequence in `signal2` for each sub-sequence in
    `signal1`. Implements the algorithm described in [Yeh2016_c]_.
    
    Parameters
    ----------
    signal1 : array
        Fisrt input time series signal.
    signal2 : array
        Second input time series signal.
    size : int
        Size of the query sub-sequences.
    index : list, array, optional
        Starting indices for query sub-sequences; the default is to search all
        sub-sequences.
    limit : int, optional
        Upper limit for the number of query sub-sequences; the default is to
        search all sub-sequences.
    
    Returns
    -------
    matrix_index : array
        Matric profile index.
    matrix_profile : array
        Computed matrix profile (distances).
    
    Notes
    -----
    * Computes euclidean distances on z-normalized data.
    
    References
    ----------
    .. [Yeh2016_c] Chin-Chia Michael Yeh, Yan Zhu, Liudmila Ulanova,
       Nurjahan Begum, Yifei Ding, Hoang Anh Dau, Diego Furtado Silva,
       Abdullah Mueen, Eamonn Keogh, "Matrix Profile I: All Pairs Similarity 
       Joins for Time Series: A Unifying View that Includes Motifs, Discords 
       and Shapelets", IEEE ICDM 2016
    
    """
    
    # check inputs
    if signal1 is None:
        raise TypeError("Please specify the first input time series signal.")
    
    if signal2 is None:
        raise TypeError("Please specify the second input time series signal.")
    
    if size is None:
        raise TypeError("Please specify the sub-sequence size.")
    
    # ensure numpy
    signal1 = np.array(signal1)
    signal2 = np.array(signal2)
    
    n1 = len(signal1)
    if size > n1/2:
        raise ValueError("First time series signal is too short relative to"
                         " desired sub-sequence length.")
    
    n2 = len(signal2)
    if size > n2/2:
        raise ValueError("Second time series signal is too short relative to"
                         " desired sub-sequence length.")
    
    if size < 4:
        raise ValueError("Sub-sequence length must be at least 4.")
    
    # matrix profile length
    nb1 = n1 - size + 1
    nb2 = n2 - size + 1
    
    # get search index
    if index is None:
        index = np.random.permutation(np.arange(nb2, dtype='int'))
    else:
        index = np.array(index)
        if not np.all(index < nb2):
            raise ValueError("Provided `index` exceeds allowable `signal2`"
                             " sub-sequences.")
    
    # limit search
    if limit is not None:
        if limit < 1:
            raise ValueError("Search limit must be at least 1.")
        
        index = index[:limit]
    
    # initialization
    matrix_profile = np.inf * np.ones(nb1, dtype='float')
    matrix_index = np.zeros(nb1, dtype='int')
    
    X, sigma = _init_dist_profile(size, n1, signal1)
    
    # compute matrix profile
    for idx in index:
        # compute distance profile
        query = signal2[idx:idx+size]
        dist = _ditance_profile(size, n1, query, X, sigma)
        dist = np.abs(np.sqrt(dist)) # to have euclidean distance
        
        # find nearest neighbor
        pos = dist <= matrix_profile
        matrix_profile[pos] = dist[pos]
        matrix_index[pos] = idx
    
    # output
    args = (matrix_index, matrix_profile)
    names = ('matrix_index', 'matrix_profile')

    return utils.ReturnTuple(args, names)


def mean_waves(data=None, size=None, step=None):
    """Extract mean samples from a data set.
    
    Parameters
    ----------
    data : array
        An m by n array of m data samples in an n-dimensional space.
    size : int
        Number of samples to use for each mean sample.
    step : int, optional
        Number of samples to jump, controlling overlap; default is equal to
        `size` (no overlap).
    
    Returns
    -------
    waves : array
        An k by n array of mean samples.
    
    Notes
    -----
    * Discards trailing samples if they are not enough to satify the `size`
      parameter.
    
    Raises
    ------
    ValueError
        If `step` is an invalid value.
    ValueError
        If there are not enough samples for the given `size`.
    
    """
    
    # check inputs
    if data is None:
        raise TypeError("Please specify an input data set.")
    
    if size is None:
        raise TypeError("Please specify the number of samples for the mean.")
    
    if step is None:
        step = size
    
    if step < 0:
        raise ValueError("The step must be a positive integer.")
    
    # number of waves
    L = len(data) - size
    nb = 1 + L // step
    if nb <= 0:
        raise ValueError("Not enough samples for the given `size`.")
    
    # compute
    waves = [np.mean(data[i:i+size], axis=0) for i in range(0, L+1, step)]
    waves = np.array(waves)
    
    return utils.ReturnTuple((waves, ), ('waves', ))


def median_waves(data=None, size=None, step=None):
    """Extract median samples from a data set.
    
    Parameters
    ----------
    data : array
        An m by n array of m data samples in an n-dimensional space.
    size : int
        Number of samples to use for each median sample.
    step : int, optional
        Number of samples to jump, controlling overlap; default is equal to
        `size` (no overlap).
    
    Returns
    -------
    waves : array
        An k by n array of median samples.
    
    Notes
    -----
    * Discards trailing samples if they are not enough to satify the `size`
      parameter.
    
    Raises
    ------
    ValueError
        If `step` is an invalid value.
    ValueError
        If there are not enough samples for the given `size`.
    
    """
    
    # check inputs
    if data is None:
        raise TypeError("Please specify an input data set.")
    
    if size is None:
        raise TypeError("Please specify the number of samples for the median.")
    
    if step is None:
        step = size
    
    if step < 0:
        raise ValueError("The step must be a positive integer.")
    
    # number of waves
    L = len(data) - size
    nb = 1 + L // step
    if nb <= 0:
        raise ValueError("Not enough samples for the given `size`.")
    
    # compute
    waves = [np.median(data[i:i+size], axis=0) for i in range(0, L+1, step)]
    waves = np.array(waves)
    
    return utils.ReturnTuple((waves, ), ('waves', ))



# -*- coding: utf-8 -*-
"""
biosppy.storage
---------------

This module provides several data storage methods.

:copyright: (c) 2015-2018 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function
from six.moves import range
import six

# built-in
import datetime
import json
import os
import zipfile

# 3rd party
import h5py
import numpy as np
import shortuuid
import joblib

# local
from . import utils


def serialize(data, path, compress=3):
    """Serialize data and save to a file using sklearn's joblib.

    Parameters
    ----------
    data : object
        Object to serialize.
    path : str
        Destination path.
    compress : int, optional
        Compression level; from 0 to 9 (highest compression).

    """

    # normalize path
    utils.normpath(path)

    joblib.dump(data, path, compress=compress)


def deserialize(path):
    """Deserialize data from a file using sklearn's joblib.

    Parameters
    ----------
    path : str
        Source path.

    Returns
    -------
    data : object
        Deserialized object.

    """

    # normalize path
    path = utils.normpath(path)

    return joblib.load(path)


def dumpJSON(data, path):
    """Save JSON data to a file.

    Parameters
    ----------
    data : dict
        The JSON data to dump.
    path : str
        Destination path.

    """

    # normalize path
    path = utils.normpath(path)

    with open(path, 'w') as fid:
        json.dump(data, fid)


def loadJSON(path):
    """Load JSON data from a file.

    Parameters
    ----------
    path : str
        Source path.

    Returns
    -------
    data : dict
        The loaded JSON data.

    """

    # normalize path
    path = utils.normpath(path)

    with open(path, 'r') as fid:
        return json.load(fid)


def zip_write(fid, files, recursive=True, root=None):
    """Write files to zip archive.

    Parameters
    ----------
    fid : file-like object
        The zip file to write into.
    files : iterable
        List of files or directories to pack.
    recursive : bool, optional
        If True, sub-directories and sub-folders are also written to the
        archive.
    root : str, optional
        Relative folder path.

    Notes
    -----
    * Ignores non-existent files and directories.

    """

    if root is None:
        root = ''

    for item in files:
        fpath = utils.normpath(item)

        if not os.path.exists(fpath):
            continue

        # relative archive name
        arcname = os.path.join(root, os.path.split(fpath)[1])

        # write
        fid.write(fpath, arcname)

        # recur
        if recursive and os.path.isdir(fpath):
            rfiles = [os.path.join(fpath, subitem)
                      for subitem in os.listdir(fpath)]
            zip_write(fid, rfiles, recursive=recursive, root=arcname)


def pack_zip(files, path, recursive=True, forceExt=True):
    """Pack files into a zip archive.

    Parameters
    ----------
    files : iterable
        List of files or directories to pack.
    path : str
        Destination path.
    recursive : bool, optional
        If True, sub-directories and sub-folders are also written to the
        archive.
    forceExt : bool, optional
        Append default extension.

    Returns
    -------
    zip_path : str
        Full path to created zip archive.

    """

    # normalize destination path
    zip_path = utils.normpath(path)

    if forceExt:
        zip_path += '.zip'

    # allowZip64 is True to allow files > 2 GB
    with zipfile.ZipFile(zip_path, 'w', allowZip64=True) as fid:
        zip_write(fid, files, recursive=recursive)

    return zip_path


def unpack_zip(zip_path, path):
    """Unpack a zip archive.

    Parameters
    ----------
    zip_path : str
        Path to zip archive.
    path : str
        Destination path (directory).

    """

    # allowZip64 is True to allow files > 2 GB
    with zipfile.ZipFile(zip_path, 'r', allowZip64=True) as fid:
        fid.extractall(path)


def alloc_h5(path):
    """Prepare an HDF5 file.

    Parameters
    ----------
    path : str
        Path to file.

    """

    # normalize path
    path = utils.normpath(path)

    with h5py.File(path):
        pass


def store_h5(path, label, data):
    """Store data to HDF5 file.

    Parameters
    ----------
    path : str
        Path to file.
    label : hashable
        Data label.
    data : array
        Data to store.

    """

    # normalize path
    path = utils.normpath(path)

    with h5py.File(path) as fid:
        label = str(label)

        try:
            fid.create_dataset(label, data=data)
        except (RuntimeError, ValueError):
            # existing label, replace
            del fid[label]
            fid.create_dataset(label, data=data)


def load_h5(path, label):
    """Load data from an HDF5 file.

    Parameters
    ----------
    path : str
        Path to file.
    label : hashable
        Data label.

    Returns
    -------
    data : array
        Loaded data.

    """

    # normalize path
    path = utils.normpath(path)

    with h5py.File(path) as fid:
        label = str(label)

        try:
            return fid[label][...]
        except KeyError:
            return None


def store_txt(path, data, sampling_rate=1000., resolution=None, date=None,
              labels=None, precision=6):
    """Store data to a simple text file.

    Parameters
    ----------
    path : str
        Path to file.
    data : array
        Data to store (up to 2 dimensions).
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    resolution : int, optional
        Sampling resolution.
    date : datetime, str, optional
        Datetime object, or an ISO 8601 formatted date-time string.
    labels : list, optional
        Labels for each column of `data`.
    precision : int, optional
        Precision for string conversion.

    Raises
    ------
    ValueError
        If the number of data dimensions is greater than 2.
    ValueError
        If the number of labels is inconsistent with the data.

    """

    # ensure numpy
    data = np.array(data)

    # check dimension
    if data.ndim > 2:
        raise ValueError("Number of data dimensions cannot be greater than 2.")

    # build header
    header = "Simple Text Format\n"
    header += "Sampling Rate (Hz):= %0.2f\n" % sampling_rate
    if resolution is not None:
        header += "Resolution:= %d\n" % resolution
    if date is not None:
        if isinstance(date, six.string_types):
            header += "Date:= %s\n" % date
        elif isinstance(date, datetime.datetime):
            header += "Date:= %s\n" % date.isoformat()
    else:
        ct = datetime.datetime.utcnow().isoformat()
        header += "Date:= %s\n" % ct

    # data type
    header += "Data Type:= %s\n" % data.dtype

    # labels
    if data.ndim == 1:
        ncols = 1
    elif data.ndim == 2:
        ncols = data.shape[1]

    if labels is None:
        labels = ['%d' % i for i in range(ncols)]
    elif len(labels) != ncols:
        raise ValueError("Inconsistent number of labels.")

    header += "Labels:= %s" % '\t'.join(labels)

    # normalize path
    path = utils.normpath(path)

    # data format
    p = '%d' % precision
    if np.issubdtype(data.dtype, np.integer):
        fmt = '%d'
    elif np.issubdtype(data.dtype, np.float):
        fmt = '%%.%sf' % p
    elif np.issubdtype(data.dtype, np.bool_):
        fmt = '%d'
    else:
        fmt = '%%.%se' % p

    # store
    np.savetxt(path, data, header=header, fmt=fmt, delimiter='\t')


def load_txt(path):
    """Load data from a text file.

    Parameters
    ----------
    path : str
        Path to file.

    Returns
    -------
    data : array
        Loaded data.
    mdata : dict
        Metadata.

    """

    # normalize path
    path = utils.normpath(path)

    with open(path, 'rb') as fid:
        lines = fid.readlines()

    # extract header
    mdata_tmp = {}
    fields = ['Sampling Rate', 'Resolution', 'Date', 'Data Type', 'Labels']
    values = []
    for item in lines:
        if b'#' in item:
            item = item.decode('utf-8')
            # parse comment
            for f in fields:
                if f in item:
                    mdata_tmp[f] = item.split(':= ')[1].strip()
                    fields.remove(f)
                    break
        else:
            values.append(item)

    # convert mdata
    mdata = {}
    df = '%Y-%m-%dT%H:%M:%S.%f'
    try:
        mdata['sampling_rate'] = float(mdata_tmp['Sampling Rate'])
    except KeyError:
        pass
    try:
        mdata['resolution'] = int(mdata_tmp['Resolution'])
    except KeyError:
        pass
    try:
        dtype = mdata_tmp['Data Type']
    except KeyError:
        dtype = None
    try:
        d = datetime.datetime.strptime(mdata_tmp['Date'], df)
        mdata['date'] = d
    except (KeyError, ValueError):
        pass
    try:
        labels = mdata_tmp['Labels'].split('\t')
        mdata['labels'] = labels
    except KeyError:
        pass

    # load array
    data = np.genfromtxt(values, dtype=dtype, delimiter=b'\t')

    return data, mdata


class HDF(object):
    """Wrapper class to operate on BioSPPy HDF5 files.

    Parameters
    ----------
    path : str
        Path to the HDF5 file.
    mode : str, optional
        File mode; one of:

        * 'a': read/write, creates file if it does not exist;
        * 'r+': read/write, file must exist;
        * 'r': read only, file must exist;
        * 'w': create file, truncate if it already exists;
        * 'w-': create file, fails if it already esists.

    """

    def __init__(self, path=None, mode='a'):
        # normalize path
        path = utils.normpath(path)

        # open file
        self._file = h5py.File(path, mode)

        # check BioSPPy structures
        try:
            self._signals = self._file['signals']
        except KeyError:
            if mode == 'r':
                raise IOError(
                    "Unable to create 'signals' group with current file mode.")
            self._signals = self._file.create_group('signals')

        try:
            self._events = self._file['events']
        except KeyError:
            if mode == 'r':
                raise IOError(
                    "Unable to create 'events' group with current file mode.")
            self._events = self._file.create_group('events')

    def __enter__(self):
        """Method for with statement."""

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Method for with statement."""

        self.close()

    def _join_group(self, *args):
        """Join group elements.

        Parameters
        ----------
        ``*args`` : list
            Group elements to join.

        Returns
        -------
        weg : str
            Joined group path.

        """

        bits = []
        for item in args:
            bits.extend(item.split('/'))

        # filter out blanks, slashes, white space
        out = []
        for item in bits:
            item = item.strip()
            if item == '':
                continue
            elif item == '/':
                continue
            out.append(item)

        weg = '/' + '/'.join(out)

        return weg

    def add_header(self, header=None):
        """Add header metadata.

        Parameters
        ----------
        header : dict
            Header metadata.

        """

        # check inputs
        if header is None:
            raise TypeError("Please specify the header information.")

        self._file.attrs['json'] = json.dumps(header)

    def get_header(self):
        """Retrieve header metadata.

        Returns
        -------
        header : dict
            Header metadata.

        """

        try:
            header = json.loads(self._file.attrs['json'])
        except KeyError:
            header = {}

        return utils.ReturnTuple((header,), ('header',))

    def add_signal(self,
                   signal=None,
                   mdata=None,
                   group='',
                   name=None,
                   compress=False):
        """Add a signal to the file.

        Parameters
        ----------
        signal : array
            Signal to add.
        mdata : dict, optional
            Signal metadata.
        group : str, optional
            Destination signal group.
        name : str, optional
            Name of the dataset to create.
        compress : bool, optional
            If True, the signal will be compressed with gzip.

        Returns
        -------
        group : str
            Destination group.
        name : str
            Name of the created signal dataset.

        """

        # check inputs
        if signal is None:
            raise TypeError("Please specify an input signal.")

        if mdata is None:
            mdata = {}

        if name is None:
            name = shortuuid.uuid()

        # navigate to group
        weg = self._join_group(self._signals.name, group)
        try:
            node = self._file[weg]
        except KeyError:
            # create group
            node = self._file.create_group(weg)

        # create dataset
        if compress:
            dset = node.create_dataset(name, data=signal, compression='gzip')
        else:
            dset = node.create_dataset(name, data=signal)

        # add metadata
        dset.attrs['json'] = json.dumps(mdata)

        # output
        grp = weg.replace('/signals', '')

        return utils.ReturnTuple((grp, name), ('group', 'name'))

    def _get_signal(self, group='', name=None):
        """Retrieve a signal dataset from the file.

        Parameters
        ----------
        group : str, optional
            Signal group.
        name : str
            Name of the signal dataset.

        Returns
        -------
        dataset : h5py.Dataset
            HDF5 dataset.

        """

        # check inputs
        if name is None:
            raise TypeError(
                "Please specify the name of the signal to retrieve.")

        # navigate to group
        weg = self._join_group(self._signals.name, group)
        try:
            node = self._file[weg]
        except KeyError:
            raise KeyError("Inexistent signal group.")

        # get data
        try:
            dset = node[name]
        except KeyError:
            raise KeyError("Inexistent signal dataset.")

        return dset

    def get_signal(self, group='', name=None):
        """Retrieve a signal from the file.

        Parameters
        ----------
        group : str, optional
            Signal group.
        name : str
            Name of the signal dataset.

        Returns
        -------
        signal : array
            Retrieved signal.
        mdata : dict
            Signal metadata.

        Notes
        -----
        * Loads the entire signal data into memory.

        """

        dset = self._get_signal(group=group, name=name)
        signal = dset[...]

        try:
            mdata = json.loads(dset.attrs['json'])
        except KeyError:
            mdata = {}

        return utils.ReturnTuple((signal, mdata), ('signal', 'mdata'))

    def del_signal(self, group='', name=None):
        """Delete a signal from the file.

        Parameters
        ----------
        group : str, optional
            Signal group.
        name : str
            Name of the dataset.

        """

        dset = self._get_signal(group=group, name=name)

        try:
            del self._file[dset.name]
        except IOError:
            raise IOError("Unable to delete object with current file mode.")

    def del_signal_group(self, group=''):
        """Delete all signals in a file group.

        Parameters
        ----------
        group : str, optional
            Signal group.

        """

        # navigate to group
        weg = self._join_group(self._signals.name, group)
        try:
            node = self._file[weg]
        except KeyError:
            raise KeyError("Inexistent signal group.")

        if node.name == '/signals':
            # delete all elements
            for _, item in six.iteritems(node):
                try:
                    del self._file[item.name]
                except IOError:
                    raise IOError(
                        "Unable to delete object with current file mode.")
        else:
            # delete single node
            try:
                del self._file[node.name]
            except IOError:
                raise IOError(
                    "Unable to delete object with current file mode.")

    def list_signals(self, group='', recursive=False):
        """List signals in the file.

        Parameters
        ----------
        group : str, optional
            Signal group.
        recursive : bool, optional
            If True, also lists signals in sub-groups.

        Returns
        -------
        signals : list
            List of (group, name) tuples of the found signals.

        """

        # navigate to group
        weg = self._join_group(self._signals.name, group)
        try:
            node = self._file[weg]
        except KeyError:
            raise KeyError("Inexistent signal group.")

        out = []
        for name, item in six.iteritems(node):
            if isinstance(item, h5py.Dataset):
                out.append((group, name))
            elif recursive and isinstance(item, h5py.Group):
                aux = self._join_group(group, name)
                out.extend(self.list_signals(group=aux,
                                             recursive=True)['signals'])

        return utils.ReturnTuple((out,), ('signals',))

    def add_event(self,
                  ts=None,
                  values=None,
                  mdata=None,
                  group='',
                  name=None,
                  compress=False):
        """Add an event to the file.

        Parameters
        ----------
        ts : array
            Array of time stamps.
        values : array, optional
            Array with data for each time stamp.
        mdata : dict, optional
            Event metadata.
        group : str, optional
            Destination event group.
        name : str, optional
            Name of the dataset to create.
        compress : bool, optional
            If True, the data will be compressed with gzip.

        Returns
        -------
        group : str
            Destination group.
        name : str
            Name of the created event dataset.

        """

        # check inputs
        if ts is None:
            raise TypeError("Please specify an input array of time stamps.")

        if values is None:
            values = []

        if mdata is None:
            mdata = {}

        if name is None:
            name = shortuuid.uuid()

        # navigate to group
        weg = self._join_group(self._events.name, group)
        try:
            node = self._file[weg]
        except KeyError:
            # create group
            node = self._file.create_group(weg)

        # create new event group
        evt_node = node.create_group(name)

        # create datasets
        if compress:
            _ = evt_node.create_dataset('ts', data=ts, compression='gzip')
            _ = evt_node.create_dataset('values',
                                        data=values,
                                        compression='gzip')
        else:
            _ = evt_node.create_dataset('ts', data=ts)
            _ = evt_node.create_dataset('values', data=values)

        # add metadata
        evt_node.attrs['json'] = json.dumps(mdata)

        # output
        grp = weg.replace('/events', '')

        return utils.ReturnTuple((grp, name), ('group', 'name'))

    def _get_event(self, group='', name=None):
        """Retrieve event datasets from the file.

        Parameters
        ----------
        group : str, optional
            Event group.
        name : str
            Name of the event dataset.

        Returns
        -------
        event : h5py.Group
            HDF5 event group.
        ts : h5py.Dataset
            HDF5 time stamps dataset.
        values : h5py.Dataset
            HDF5 values dataset.

        """

        # check inputs
        if name is None:
            raise TypeError(
                "Please specify the name of the signal to retrieve.")

        # navigate to group
        weg = self._join_group(self._events.name, group)
        try:
            node = self._file[weg]
        except KeyError:
            raise KeyError("Inexistent event group.")

        # event group
        try:
            evt_group = node[name]
        except KeyError:
            raise KeyError("Inexistent event dataset.")

        # get data
        try:
            ts = evt_group['ts']
        except KeyError:
            raise KeyError("Could not find expected time stamps structure.")

        try:
            values = evt_group['values']
        except KeyError:
            raise KeyError("Could not find expected values structure.")

        return evt_group, ts, values

    def get_event(self, group='', name=None):
        """Retrieve an event from the file.

        Parameters
        ----------
        group : str, optional
            Event group.
        name : str
            Name of the event dataset.

        Returns
        -------
        ts : array
            Array of time stamps.
        values : array
            Array with data for each time stamp.
        mdata : dict
            Event metadata.

        Notes
        -----
        Loads the entire event data into memory.

        """

        evt_group, dset_ts, dset_values = self._get_event(group=group,
                                                          name=name)
        ts = dset_ts[...]
        values = dset_values[...]

        try:
            mdata = json.loads(evt_group.attrs['json'])
        except KeyError:
            mdata = {}

        return utils.ReturnTuple((ts, values, mdata),
                                 ('ts', 'values', 'mdata'))

    def del_event(self, group='', name=None):
        """Delete an event from the file.

        Parameters
        ----------
        group : str, optional
            Event group.
        name : str
            Name of the event dataset.

        """

        evt_group, _, _ = self._get_event(group=group, name=name)

        try:
            del self._file[evt_group.name]
        except IOError:
            raise IOError("Unable to delete object with current file mode.")

    def del_event_group(self, group=''):
        """Delete all events in a file group.

        Parameters
        ----------
        group  str, optional
            Event group.

        """

        # navigate to group
        weg = self._join_group(self._events.name, group)
        try:
            node = self._file[weg]
        except KeyError:
            raise KeyError("Inexistent event group.")

        if node.name == '/events':
            # delete all elements
            for _, item in six.iteritems(node):
                try:
                    del self._file[item.name]
                except IOError:
                    raise IOError(
                        "Unable to delete object with current file mode.")
        else:
            # delete single node
            try:
                del self._file[node.name]
            except IOError:
                raise IOError(
                    "Unable to delete object with current file mode.")

    def list_events(self, group='', recursive=False):
        """List events in the file.

        Parameters
        ----------
        group : str, optional
            Event group.
        recursive : bool, optional
            If True, also lists events in sub-groups.

        Returns
        -------
        events : list
            List of (group, name) tuples of the found events.

        """

        # navigate to group
        weg = self._join_group(self._events.name, group)
        try:
            node = self._file[weg]
        except KeyError:
            raise KeyError("Inexistent event group.")

        out = []
        for name, item in six.iteritems(node):
            if isinstance(item, h5py.Group):
                try:
                    _ = item.attrs['json']
                except KeyError:
                    # normal group
                    if recursive:
                        aux = self._join_group(group, name)
                        out.extend(self.list_events(group=aux,
                                                    recursive=True)['events'])
                else:
                    # event group
                    out.append((group, name))

        return utils.ReturnTuple((out,), ('events',))

    def close(self):
        """Close file descriptor."""

        # flush buffers
        self._file.flush()

        # close
        self._file.close()



# -*- coding: utf-8 -*-
"""
biosppy.metrics
---------------

This module provides pairwise distance computation methods.

:copyright: (c) 2015-2018 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function
import six

# 3rd party
import numpy as np
import scipy.spatial.distance as ssd
from scipy import linalg


def pcosine(u, v):
    """Computes the Cosine distance (positive space) between 1-D arrays.

    The Cosine distance (positive space) between `u` and `v` is defined as

    .. math::

        d(u, v) = 1 - abs \\left( \\frac{u \\cdot v}{||u||_2 ||v||_2} \\right)

    where :math:`u \\cdot v` is the dot product of :math:`u` and :math:`v`.

    Parameters
    ----------
    u : array
        Input array.
    v : array
        Input array.

    Returns
    -------
    cosine : float
        Cosine distance between `u` and `v`.

    """

    # validate vectors like scipy does
    u = ssd._validate_vector(u)
    v = ssd._validate_vector(v)

    dist = 1. - np.abs(np.dot(u, v) / (linalg.norm(u) * linalg.norm(v)))

    return dist


def pdist(X, metric='euclidean', p=2, w=None, V=None, VI=None):
    """Pairwise distances between observations in n-dimensional space.

    Wraps scipy.spatial.distance.pdist.

    Parameters
    ----------
    X : array
        An m by n array of m original observations in an n-dimensional space.
    metric : str, function, optional
        The distance metric to use; the distance can be 'braycurtis',
        'canberra', 'chebyshev', 'cityblock', 'correlation', 'cosine', 'dice',
        'euclidean', 'hamming', 'jaccard', 'kulsinski', 'mahalanobis',
        'matching', 'minkowski', 'pcosine', 'rogerstanimoto', 'russellrao',
        'seuclidean', 'sokalmichener', 'sokalsneath', 'sqeuclidean', 'yule'.
    p : float, optional
        The p-norm to apply (for Minkowski, weighted and unweighted).
    w : array, optional
        The weight vector (for weighted Minkowski).
    V : array, optional
        The variance vector (for standardized Euclidean).
    VI : array, optional
        The inverse of the covariance matrix (for Mahalanobis).

    Returns
    -------
    Y : array
        Returns a condensed distance matrix Y.  For each :math:`i` and
        :math:`j` (where :math:`i<j<n`), the metric ``dist(u=X[i], v=X[j])``
        is computed and stored in entry ``ij``.

    """

    if isinstance(metric, six.string_types):
        if metric == 'pcosine':
            metric = pcosine

    return ssd.pdist(X, metric, p, w, V, VI)


def cdist(XA, XB, metric='euclidean', p=2, V=None, VI=None, w=None):
    """Computes distance between each pair of the two collections of inputs.

    Wraps scipy.spatial.distance.cdist.

    Parameters
    ----------
    XA : array
        An :math:`m_A` by :math:`n` array of :math:`m_A` original observations
        in an :math:`n`-dimensional space.
    XB : array
        An :math:`m_B` by :math:`n` array of :math:`m_B` original observations
        in an :math:`n`-dimensional space.
    metric : str, function, optional
        The distance metric to use; the distance can be 'braycurtis',
        'canberra', 'chebyshev', 'cityblock', 'correlation', 'cosine', 'dice',
        'euclidean', 'hamming', 'jaccard', 'kulsinski', 'mahalanobis',
        'matching', 'minkowski', 'pcosine', 'rogerstanimoto', 'russellrao',
        'seuclidean', 'sokalmichener', 'sokalsneath', 'sqeuclidean', 'yule'.
    p : float, optional
        The p-norm to apply (for Minkowski, weighted and unweighted).
    w : array, optional
        The weight vector (for weighted Minkowski).
    V : array, optional
        The variance vector (for standardized Euclidean).
    VI : array, optional
        The inverse of the covariance matrix (for Mahalanobis).

    Returns
    -------
    Y : array
        An :math:`m_A` by :math:`m_B` distance matrix is returned. For each
        :math:`i` and :math:`j`, the metric ``dist(u=XA[i], v=XB[j])``
        is computed and stored in the :math:`ij` th entry.

    """

    if isinstance(metric, six.string_types):
        if metric == 'pcosine':
            metric = pcosine

    return ssd.cdist(XA, XB, metric, p, V, VI, w)


def squareform(X, force="no", checks=True):
    """Converts a vector-form distance vector to a square-form distance matrix,
    and vice-versa.

    Wraps scipy.spatial.distance.squareform.

    Parameters
    ----------
    X : array
        Either a condensed or redundant distance matrix.
    force : str, optional
        As with MATLAB(TM), if force is equal to 'tovector' or 'tomatrix', the
        input will be treated as a distance matrix or distance vector
        respectively.
    checks : bool, optional
        If `checks` is set to False, no checks will be made for matrix
        symmetry nor zero diagonals. This is useful if it is known that
        ``X - X.T1`` is small and ``diag(X)`` is close to zero. These values
        are ignored any way so they do not disrupt the squareform
        transformation.

    Returns
    -------
    Y : array
        If a condensed distance matrix is passed, a redundant one is returned,
        or if a redundant one is passed, a condensed distance matrix is
        returned.

    """

    return ssd.squareform(X, force, checks)



# -*- coding: utf-8 -*-
"""
biosppy.signals.resp
--------------------

This module provides methods to process Respiration (Resp) signals.

:copyright: (c) 2015-2018 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function

# 3rd party
import numpy as np

# local
from . import tools as st
from .. import plotting, utils


def resp(signal=None, sampling_rate=1000., path=None, show=True):
    """Process a raw Respiration signal and extract relevant signal features
    using default parameters.

    Parameters
    ----------
    signal : array
        Raw Respiration signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show a summary plot.

    Returns
    -------
    ts : array
        Signal time axis reference (seconds).
    filtered : array
        Filtered Respiration signal.
    zeros : array
        Indices of Respiration zero crossings.
    resp_rate_ts : array
        Respiration rate time axis reference (seconds).
    resp_rate : array
        Instantaneous respiration rate (Hz).

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    sampling_rate = float(sampling_rate)

    # filter signal
    filtered, _, _ = st.filter_signal(signal=signal,
                                      ftype='butter',
                                      band='bandpass',
                                      order=2,
                                      frequency=[0.1, 0.35],
                                      sampling_rate=sampling_rate)

    # compute zero crossings
    zeros, = st.zero_cross(signal=filtered, detrend=True)
    beats = zeros[::2]

    if len(beats) < 2:
        rate_idx = []
        rate = []
    else:
        # compute respiration rate
        rate_idx = beats[1:]
        rate = sampling_rate * (1. / np.diff(beats))

        # physiological limits
        indx = np.nonzero(rate <= 0.35)
        rate_idx = rate_idx[indx]
        rate = rate[indx]

        # smooth with moving average
        size = 3
        rate, _ = st.smoother(signal=rate,
                              kernel='boxcar',
                              size=size,
                              mirror=True)

    # get time vectors
    length = len(signal)
    T = (length - 1) / sampling_rate
    ts = np.linspace(0, T, length, endpoint=True)
    ts_rate = ts[rate_idx]

    # plot
    if show:
        plotting.plot_resp(ts=ts,
                           raw=signal,
                           filtered=filtered,
                           zeros=zeros,
                           resp_rate_ts=ts_rate,
                           resp_rate=rate,
                           path=path,
                           show=True)

    # output
    args = (ts, filtered, zeros, ts_rate, rate)
    names = ('ts', 'filtered', 'zeros', 'resp_rate_ts', 'resp_rate')

    return utils.ReturnTuple(args, names)



# -*- coding: utf-8 -*-
"""
biosppy.biometrics
------------------

This module provides classifier interfaces for identity recognition
(biometrics) applications. The core API methods are:
* enroll: add a new subject;
* dismiss: remove an existing subject;
* identify: determine the identity of collected biometric dataset;
* authenticate: verify the identity of collected biometric dataset.

:copyright: (c) 2015-2018 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function
from six.moves import range
import six

# built-in
import collections

# 3rd party
import numpy as np
import shortuuid
from bidict import bidict
from sklearn import model_selection as skcv
from sklearn import svm as sksvm

# local
from . import metrics, plotting, storage, utils
from .signals import tools


class SubjectError(Exception):
    """Exception raised when the subject is unknown."""

    def __init__(self, subject=None):
        self.subject = subject

    def __str__(self):
        if self.subject is None:
            return str("Subject is not enrolled.")
        else:
            return str("Subject %r is not enrolled." % self.subject)


class UntrainedError(Exception):
    """Exception raised when classifier is not trained."""

    def __str__(self):
        return str("The classifier is not trained.")


class CombinationError(Exception):
    """Exception raised when the combination method fails."""

    def __str__(self):
        return str("Combination of empty array.")


class BaseClassifier(object):
    """Base biometric classifier class.

    This class is a skeleton for actual classifier classes.
    The following methods must be overridden or adapted to build a
    new classifier:
    
    * __init__
    * _authenticate
    * _get_thresholds
    * _identify
    * _prepare
    * _train
    * _update

    Attributes
    ----------
    EER_IDX : int
        Reference index for the Equal Error Rate.

    """

    EER_IDX = 0

    def __init__(self):
        # generic self things
        self.is_trained = False
        self._subject2label = bidict()
        self._nbSubjects = 0
        self._thresholds = {}
        self._autoThresholds = None

        # init data storage
        self._iofile = {}

        # defer flag
        self._defer_flag = False
        self._reset_defer()

    def _reset_defer(self):
        """Reset defer buffer."""

        self._defer_dict = {'enroll': set(), 'dismiss': set()}

    def _defer(self, label, case):
        """Add deferred task.

        Parameters
        ----------
        label : str
            Internal classifier subject label.
        case : str
            One of 'enroll' or 'dismiss'.

        Notes
        -----
        * An enroll overrides a previous dismiss for the same subject.
        * A dismiss overrides a previous enroll for the same subject.

        """

        if case == 'enroll':
            self._defer_dict['enroll'].add(label)
            if label in self._defer_dict['dismiss']:
                self._defer_dict['dismiss'].remove(label)
        elif case == 'dismiss':
            self._defer_dict['dismiss'].add(label)
            if label in self._defer_dict['enroll']:
                self._defer_dict['enroll'].remove(label)

        self._defer_flag = True

    def _check_state(self):
        """Check and update the train state."""

        if self._nbSubjects > 0:
            self.is_trained = True
        else:
            self.is_trained = False

    def io_load(self, label):
        """Load enrolled subject data.

        Parameters
        ----------
        label : str
            Internal classifier subject label.

        Returns
        -------
        data : array
            Subject data.

        """

        return self._iofile[label]

    def io_save(self, label, data):
        """Save subject data.

        Parameters
        ----------
        label : str
            Internal classifier subject label.
        data : array
            Subject data.

        """

        self._iofile[label] = data

    def io_del(self, label):
        """Delete subject data.

        Parameters
        ----------
        label : str
            Internal classifier subject label.

        """

        del self._iofile[label]

    def save(self, path):
        """Save classifier instance to a file.

        Parameters
        ----------
        path : str
            Destination file path.

        """

        storage.serialize(self, path)

    @classmethod
    def load(cls, path):
        """Load classifier instance from a file.

        Parameters
        ----------
        path : str
            Source file path.

        Returns
        -------
        clf : object
            Loaded classifier instance.

        """

        # load classifier
        clf = storage.deserialize(path)

        # check class type
        if not isinstance(clf, cls):
            raise TypeError("Mismatch between target class and loaded file.")

        return clf

    def check_subject(self, subject):
        """Check if a subject is enrolled.

        Parameters
        ----------
        subject : hashable
            Subject identity.

        Returns
        -------
        check : bool
            If True, the subject is enrolled.

        """

        if self.is_trained:
            return subject in self._subject2label

        return False

    def list_subjects(self):
        """List all the enrolled subjects.

        Returns
        -------
        subjects : list
            Enrolled subjects.

        """

        subjects = list(self._subject2label)

        return subjects

    def enroll(self, data=None, subject=None, deferred=False):
        """Enroll new data for a subject.

        If the subject is already enrolled, new data is combined with
        existing data.

        Parameters
        ----------
        data : array
            Data to enroll.
        subject : hashable
            Subject identity.
        deferred : bool, optional
            If True, computations are delayed until `flush` is called.

        Notes
        -----
        * When using deferred calls, an enroll overrides a previous dismiss
          for the same subject.

        """

        # check inputs
        if data is None:
            raise TypeError("Please specify the data to enroll.")

        if subject is None:
            raise TypeError("Plase specify the subject identity.")

        if self.check_subject(subject):
            # load existing
            label = self._subject2label[subject]
            old = self.io_load(label)

            # combine data
            data = self._update(old, data)
        else:
            # create new label
            label = shortuuid.uuid()
            self._subject2label[subject] = label
            self._nbSubjects += 1

        # store data
        self.io_save(label, data)

        if deferred:
            # delay computations
            self._defer(label, 'enroll')
        else:
            self._train([label], None)
            self._check_state()
            self.update_thresholds()

    def dismiss(self, subject=None, deferred=False):
        """Remove a subject.

        Parameters
        ----------
        subject : hashable
            Subject identity.
        deferred : bool, optional
            If True, computations are delayed until `flush` is called.

        Raises
        ------
        SubjectError
            If the subject to remove is not enrolled.

        Notes
        -----
        * When using deferred calls, a dismiss overrides a previous enroll
          for the same subject.

        """

        # check inputs
        if subject is None:
            raise TypeError("Please specify the subject identity.")

        if not self.check_subject(subject):
            raise SubjectError(subject)

        label = self._subject2label[subject]
        del self._subject2label[subject]
        del self._thresholds[label]
        self._nbSubjects -= 1
        self.io_del(label)

        if deferred:
            self._defer(label, 'dismiss')
        else:
            self._train(None, [label])
            self._check_state()
            self.update_thresholds()

    def batch_train(self, data=None):
        """Train the classifier in batch mode.

        Parameters
        ----------
        data : dict
            Dictionary holding training data for each subject; if the object
            for a subject is `None`, performs a `dismiss`.

        """

        # check inputs
        if data is None:
            raise TypeError("Please specify the data to train.")

        for sub, val in six.iteritems(data):
            if val is None:
                try:
                    self.dismiss(sub, deferred=True)
                except SubjectError:
                    continue
            else:
                self.enroll(val, sub, deferred=True)

        self.flush()

    def flush(self):
        """Flush deferred computations."""

        if self._defer_flag:
            self._defer_flag = False

            # train
            enroll = list(self._defer_dict['enroll'])
            dismiss = list(self._defer_dict['dismiss'])
            self._train(enroll, dismiss)

            # update thresholds
            self._check_state()
            self.update_thresholds()

            # reset
            self._reset_defer()

    def update_thresholds(self, fraction=1.):
        """Update subject-specific thresholds based on the enrolled data.

        Parameters
        ----------
        fraction : float, optional
            Fraction of samples to select from training data.

        """

        ths = self.get_thresholds(force=True)

        # gather data to test
        data = {}
        for subject, label in six.iteritems(self._subject2label):
            # select a random fraction of the training data
            aux = self.io_load(label)
            indx = list(range(len(aux)))
            use, _ = utils.random_fraction(indx, fraction, sort=True)

            data[subject] = aux[use]

        # evaluate classifier
        _, res = self.evaluate(data, ths)

        # choose thresholds at EER
        for subject, label in six.iteritems(self._subject2label):
            EER_auth = res['subject'][subject]['authentication']['rates']['EER']
            self.set_auth_thr(label, EER_auth[self.EER_IDX, 0], ready=True)

            EER_id = res['subject'][subject]['identification']['rates']['EER']
            self.set_id_thr(label, EER_id[self.EER_IDX, 0], ready=True)

    def set_auth_thr(self, subject, threshold, ready=False):
        """Set the authentication threshold of a subject.

        Parameters
        ----------
        subject : hashable
            Subject identity.
        threshold : int, float
            Threshold value.
        ready : bool, optional
            If True, `subject` is the internal classifier label.

        """

        if not ready:
            if not self.check_subject(subject):
                raise SubjectError(subject)
            subject = self._subject2label[subject]

        try:
            self._thresholds[subject]['auth'] = threshold
        except KeyError:
            self._thresholds[subject] = {'auth': threshold, 'id': None}

    def get_auth_thr(self, subject, ready=False):
        """Get the authentication threshold of a subject.

        Parameters
        ----------
        subject : hashable
            Subject identity.
        ready : bool, optional
            If True, `subject` is the internal classifier label.

        Returns
        -------
        threshold : int, float
            Threshold value.

        """

        if not ready:
            if not self.check_subject(subject):
                raise SubjectError(subject)
            subject = self._subject2label[subject]

        return self._thresholds[subject].get('auth', None)

    def set_id_thr(self, subject, threshold, ready=False):
        """Set the identification threshold of a subject.

        Parameters
        ----------
        subject : hashable
            Subject identity.
        threshold : int, float
            Threshold value.
        ready : bool, optional
            If True, `subject` is the internal classifier label.

        """

        if not ready:
            if not self.check_subject(subject):
                raise SubjectError(subject)
            subject = self._subject2label[subject]

        try:
            self._thresholds[subject]['id'] = threshold
        except KeyError:
            self._thresholds[subject] = {'auth': None, 'id': threshold}

    def get_id_thr(self, subject, ready=False):
        """Get the identification threshold of a subject.

        Parameters
        ----------
        subject : hashable
            Subject identity.
        ready : bool, optional
            If True, `subject` is the internal classifier label.

        Returns
        -------
        threshold : int, float
            Threshold value.

        """

        if not ready:
            if not self.check_subject(subject):
                raise SubjectError(subject)
            subject = self._subject2label[subject]

        return self._thresholds[subject].get('id', None)

    def get_thresholds(self, force=False):
        """Get an array of reasonable thresholds.

        Parameters
        ----------
        force : bool, optional
            If True, forces generation of thresholds.

        Returns
        -------
        ths : array
            Generated thresholds.

        """

        if force or (self._autoThresholds is None):
            self._autoThresholds = self._get_thresholds()

        return self._autoThresholds

    def authenticate(self, data, subject, threshold=None):
        """Authenticate a set of feature vectors, allegedly belonging to the
        given subject.

        Parameters
        ----------
        data : array
            Input test data.
        subject : hashable
            Subject identity.
        threshold : int, float, optional
            Authentication threshold.

        Returns
        -------
        decision : array
            Authentication decision for each input sample.

        """

        # check train state
        if not self.is_trained:
            raise UntrainedError

        # check subject
        if not self.check_subject(subject):
            raise SubjectError(subject)

        label = self._subject2label[subject]

        # check threshold
        if threshold is None:
            threshold = self.get_auth_thr(label, ready=True)

        # prepare data
        aux = self._prepare(data, targets=label)

        # authenticate
        decision = self._authenticate(aux, label, threshold)

        return decision

    def identify(self, data, threshold=None):
        """Identify a set of feature vectors.

        Parameters
        ----------
        data : array
            Input test data.
        threshold : int, float, optional
            Identification threshold.

        Returns
        -------
        subjects : list
            Identity of each input sample.

        """

        # check train state
        if not self.is_trained:
            raise UntrainedError

        # prepare data
        aux = self._prepare(data)

        # identify
        labels = self._identify(aux, threshold)

        # translate class labels
        subjects = [self._subject2label.inv.get(item, '') for item in labels]

        return subjects

    def evaluate(self, data, thresholds=None, path=None, show=False):
        """Assess the performance of the classifier in both authentication and
        identification scenarios.

        Parameters
        ----------
        data : dict
            Dictionary holding test data for each subject.
        thresholds : array, optional
            Classifier thresholds to use.
        path : str, optional
            If provided, the plot will be saved to the specified file.
        show : bool, optional
            If True, show a summary plot.

        Returns
        -------
        classification : dict
            Classification results.
        assessment : dict
            Biometric statistics.

        """

        # check train state
        if not self.is_trained:
            raise UntrainedError

        # check thresholds
        if thresholds is None:
            thresholds = self.get_thresholds()

        # get subjects
        subjects = [item for item in data if self.check_subject(item)]
        if len(subjects) == 0:
            raise ValueError("No enrolled subjects in test set.")

        results = {
            'subjectList': subjects,
            'subjectDict': self._subject2label,
        }

        for subject in subjects:
            # prepare data
            aux = self._prepare(data[subject])

            # test
            auth_res = []
            id_res = []
            for th in thresholds:
                # authentication
                auth = []
                for subject_tst in subjects:
                    label = self._subject2label[subject_tst]
                    auth.append(self._authenticate(aux, label, th))

                auth_res.append(np.array(auth))

                # identification
                id_res.append(self._identify(aux, th))

            auth_res = np.array(auth_res)
            id_res = np.array(id_res)
            results[subject] = {'authentication': auth_res,
                                'identification': id_res,
                                }

        # assess classification results
        assess, = assess_classification(results, thresholds)

        # output
        args = (results, assess)
        names = ('classification', 'assessment')
        out = utils.ReturnTuple(args, names)

        if show:
            # plot
            plotting.plot_biometrics(assess,
                                     self.EER_IDX,
                                     path=path,
                                     show=True)

        return out

    @classmethod
    def cross_validation(cls, data, labels, cv, thresholds=None, **kwargs):
        """Perform Cross Validation (CV) on a data set.

        Parameters
        ----------
        data : array
            An m by n array of m data samples in an n-dimensional space.
        labels : list, array
            A list of m class labels.
        cv : CV iterator
            A `sklearn.model_selection` iterator.
        thresholds : array, optional
            Classifier thresholds to use.
        ``**kwargs`` : dict, optional
            Classifier parameters.

        Returns
        -------
        runs : list
            Evaluation results for each CV run.
        assessment : dict
            Final CV biometric statistics.

        """

        runs = []
        aux = []
        for train, test in cv:
            # train data set
            train_idx = collections.defaultdict(list)
            for item in train:
                lbl = labels[item]
                train_idx[lbl].append(item)

            train_data = {sub: data[idx] for sub, idx in six.iteritems(train_idx)}

            # test data set
            test_idx = collections.defaultdict(list)
            for item in test:
                lbl = labels[item]
                test_idx[lbl].append(item)

            test_data = {sub: data[idx] for sub, idx in six.iteritems(test_idx)}

            # instantiate classifier
            clf = cls(**kwargs)
            clf.batch_train(train_data)
            res = clf.evaluate(test_data, thresholds=thresholds)
            del clf

            aux.append(res['assessment'])
            runs.append(res)

        # assess runs
        if len(runs) > 0:
            subjects = runs[0]['classification']['subjectList']
            assess, = assess_runs(results=aux, subjects=subjects)
        else:
            raise ValueError("CV iterator empty or exhausted.")

        # output
        args = (runs, assess)
        names = ('runs', 'assessment')

        return utils.ReturnTuple(args, names)

    def _authenticate(self, data, label, threshold):
        """Authenticate a set of feature vectors, allegedly belonging to the
        given subject.

        Parameters
        ----------
        data : array
            Input test data.
        label : str
            Internal classifier subject label.
        threshold : int, float
            Authentication threshold.

        Returns
        -------
        decision : array
            Authentication decision for each input sample.

        """

        decision = np.zeros(len(data), dtype='bool')

        return decision

    def _get_thresholds(self):
        """Generate an array of reasonable thresholds.

        Returns
        -------
        ths : array
            Generated thresholds.

        """

        ths = np.array([])

        return ths

    def _identify(self, data, threshold=None):
        """Identify a set of feature vectors.

        Parameters
        ----------
        data : array
            Input test data.
        threshold : int, float
            Identification threshold.

        Returns
        -------
        labels : list
            Identity (internal label) of each input sample.

        """

        labels = [''] * len(data)

        return labels

    def _prepare(self, data, targets=None):
        """Prepare data to be processed.

        Parameters
        ----------
        data : array
            Data to process.
        targets : list, str, optional
            Target subject labels.

        Returns
        -------
        out : object
            Processed data.

        """

        # target class labels
        if targets is None:
            targets = list(self._subject2label.values())
        elif isinstance(targets, six.string_types):
            targets = [targets]

        return data

    def _train(self, enroll=None, dismiss=None):
        """Train the classifier.

        Parameters
        ----------
        enroll : list, optional
            Labels of new or updated subjects.
        dismiss : list, optional
            Labels of deleted subjects.

        """

        if enroll is None:
            enroll = []
        if dismiss is None:
            dismiss = []

        # process dismiss
        for _ in dismiss:
            pass

        # process enroll
        for _ in enroll:
            pass

    def _update(self, old, new):
        """Combine new data with existing templates (for one subject).

        Parameters
        ----------
        old : array
            Existing data.
        new : array
            New data.

        Returns
        -------
        out : array
            Combined data.

        """

        return new


class KNN(BaseClassifier):
    """K Nearest Neighbors (k-NN) biometric classifier.

    Parameters
    ----------
    k : int, optional
        Number of neighbors.
    metric : str, optional
        Distance metric.
    metric_args : dict, optional
        Additional keyword arguments are passed to the distance function.

    Attributes
    ----------
    EER_IDX : int
        Reference index for the Equal Error Rate.

    """

    EER_IDX = 0

    def __init__(self, k=3, metric='euclidean', metric_args=None):
        # parent __init__
        super(KNN, self).__init__()

        # algorithm self things
        self.k = k
        self.metric = metric
        if metric_args is None:
            metric_args = {}
        self.metric_args = metric_args

        # test metric args
        _ = metrics.pdist(np.zeros((2, 2)), metric, **metric_args)

        # minimum threshold
        self.min_thr = 10 * np.finfo('float').eps

    def _sort(self, dists, train_labels):
        """Sort the computed distances.

        Parameters
        ----------
        dists : array
            Unsorted computed distances.
        train_labels : list
            Unsorted target subject labels.

        Returns
        -------
        dists : array
            Sorted computed distances.
        train_labels : list
            Sorted target subject labels.

        """

        ind = dists.argsort()
        # sneaky trick from http://stackoverflow.com/questions/6155649
        static_inds = np.arange(dists.shape[0]).reshape((dists.shape[0], 1))
        dists = dists[static_inds, ind]
        train_labels = train_labels[static_inds, ind]

        return dists, train_labels

    def _authenticate(self, data, label, threshold):
        """Authenticate a set of feature vectors, allegedly belonging to the
        given subject.

        Parameters
        ----------
        data : array
            Input test data.
        label : str
            Internal classifier subject label.
        threshold : int, float
            Authentication threshold.

        Returns
        -------
        decision : array
            Authentication decision for each input sample.

        """

        # unpack prepared data
        dists = data['dists']
        train_labels = data['train_labels']

        # select based on subject label
        aux = []
        ns = len(dists)
        for i in range(ns):
            aux.append(dists[i, train_labels[i, :] == label])

        dists = np.array(aux)

        # nearest neighbors
        dists = dists[:, :self.k]

        decision = np.zeros(ns, dtype='bool')
        for i in range(ns):
            # compare distances to threshold
            count = np.sum(dists[i, :] <= threshold)

            # decide accept
            if count > (self.k // 2):
                decision[i] = True

        return decision

    def _get_thresholds(self):
        """Generate an array of reasonable thresholds.

        For metrics other than 'cosine' or 'pcosine', which have a clear
        limits, generates an array based on the maximum distances between
        enrolled subjects.

        Returns
        -------
        ths : array
            Generated thresholds.

        """

        if self.metric == 'cosine':
            return np.linspace(self.min_thr, 2., 100)
        elif self.metric == 'pcosine':
            return np.linspace(self.min_thr, 1., 100)

        maxD = []
        for _ in range(3):
            for label in list(six.itervalues(self._subject2label)):
                # randomly select samples
                aux = self.io_load(label)
                ind = np.random.randint(0, aux.shape[0], 3)
                obs = aux[ind]

                # compute distances
                dists = self._prepare(obs)['dists']
                maxD.append(np.max(dists))

        # maximum distance
        maxD = 1.5 * np.max(maxD)

        ths = np.linspace(self.min_thr, maxD, 100)

        return ths

    def _identify(self, data, threshold=None):
        """Identify a set of feature vectors.

        Parameters
        ----------
        data : array
            Input test data.
        threshold : int, float
            Identification threshold.

        Returns
        -------
        labels :list
            Identity (internal label) of each input sample.

        """

        if threshold is None:
            thrFcn = lambda label: self.get_id_thr(label, ready=True)
        else:
            thrFcn = lambda label: threshold

        # unpack prepared data
        dists = data['dists']
        train_labels = data['train_labels']

        # nearest neighbors
        dists = dists[:, :self.k]
        train_labels = train_labels[:, :self.k]
        ns = len(dists)

        labels = []
        for i in range(ns):
            lbl, _ = majority_rule(train_labels[i, :], random=True)

            # compare distances to threshold
            count = np.sum(dists[i, :] <= thrFcn(lbl))

            # decide
            if count > (self.k // 2):
                # accept
                labels.append(lbl)
            else:
                # reject
                labels.append('')

        return labels

    def _prepare(self, data, targets=None):
        """Prepare data to be processed.

        Computes the distances of the input data set to the target subjects.

        Parameters
        ----------
        data : array
            Data to process.
        targets : list, str, optional
            Target subject labels.

        Returns
        -------
        out : dict
            Processed data containing the computed distances (`dists`) and the
            target subject labels (`train_labels`).

        """

        # target class labels
        if targets is None:
            targets = list(six.itervalues(self._subject2label))
        elif isinstance(targets, six.string_types):
            targets = [targets]

        dists = []
        train_labels = []

        for label in targets:
            # compute distances
            D = metrics.cdist(data, self.io_load(label),
                              metric=self.metric, **self.metric_args)

            dists.append(D)
            train_labels.append(np.tile(label, D.shape))

        dists = np.concatenate(dists, axis=1)
        train_labels = np.concatenate(train_labels, axis=1)

        # sort
        dists, train_labels = self._sort(dists, train_labels)

        return {'dists': dists, 'train_labels': train_labels}

    def _update(self, old, new):
        """Combine new data with existing templates (for one subject).

        Simply concatenates old data with new data.

        Parameters
        ----------
        old : array
            Existing data.
        new : array
            New data.

        Returns
        -------
        out : array
            Combined data.

        """

        out = np.concatenate([old, new], axis=0)

        return out


class SVM(BaseClassifier):
    """Support Vector Machines (SVM) biometric classifier.

    Wraps the 'OneClassSVM' and 'SVC' classes from 'scikit-learn'.

    Parameters
    ----------
    C : float, optional
        Penalty parameter C of the error term.
    kernel : str, optional
        Specifies the kernel type to be used in the algorithm. It must be one
        of ‘linear’, ‘poly’, ‘rbf’, ‘sigmoid’, ‘precomputed’ or a callable.
        If none is given, ‘rbf’ will be used. If a callable is given it is
        used to precompute the kernel matrix.
    degree : int, optional
        Degree of the polynomial kernel function (‘poly’). Ignored by all other
        kernels.
    gamma : float, optional
        Kernel coefficient for ‘rbf’, ‘poly’ and ‘sigmoid’. If gamma is 'auto'
        then 1/n_features will be used instead.
    coef0 : float, optional
        Independent term in kernel function. It is only significant in ‘poly’
        and ‘sigmoid’.
    shrinking : bool, optional
        Whether to use the shrinking heuristic.
    tol : float, optional
        Tolerance for stopping criterion.
    cache_size : float, optional
        Specify the size of the kernel cache (in MB).
    max_iter : int, optional
        Hard limit on iterations within solver, or -1 for no limit.
    random_state : int, RandomState, optional
        The seed of the pseudo random number generator to use when shuffling
        the data for probability estimation.

    Attributes
    ----------
    EER_IDX : int
        Reference index for the Equal Error Rate.

    """

    EER_IDX = -1

    def __init__(self,
                 C=1.0,
                 kernel='linear',
                 degree=3,
                 gamma='auto',
                 coef0=0.0,
                 shrinking=True,
                 tol=0.001,
                 cache_size=200,
                 max_iter=-1,
                 random_state=None):
        # parent __init__
        super(SVM, self).__init__()

        # algorithm self things
        self._models = {}
        self._clf_kwargs = {
            'C': C,
            'kernel': kernel,
            'degree': degree,
            'gamma': gamma,
            'coef0': coef0,
            'shrinking': shrinking,
            'tol': tol,
            'cache_size': cache_size,
            'max_iter': max_iter,
            'random_state': random_state,
        }

        # minimum threshold
        self.min_thr = 10 * np.finfo('float').eps

    def _get_weights(self, n1, n2):
        """Compute class weights.

        The weights are inversely proportional to the number of samples in each
        class.

        Parameters
        ----------
        n1 : int
            Number of samples in the first class.
        n2 : int
            Number of samples in the second class.

        Returns
        -------
        weights : dict
            Weights for each class.

        """

        w = np.array([1. / n1, 1. / n2])
        w *= 2 / np.sum(w)
        weights = {-1: w[0], 1: w[1]}

        return weights

    def _get_single_clf(self, X, label):
        """Instantiate and train a One Class SVM classifier.

        Parameters
        ----------
        X : array
            Training data.
        label : str
            Class label.

        """

        clf = sksvm.OneClassSVM(kernel='rbf', nu=0.1)
        clf.fit(X)

        # add to models
        self._models[('', label)] = clf

    def _get_kernel_clf(self, X1, X2, n1, n2, label1, label2):
        """Instantiate and train a SVC SVM classifier.

        Parameters
        ----------
        X1 : array
            Trainig data for the first class.
        X2 : array
            Training data for the second class.
        n1 : int
            Number of samples in the first class.
        n2 : int
            Number of samples in the second class.
        label1 : str
            Label for the first class.
        label2 : str
            Label for the first class.

        """

        # prepare data to train
        X = np.concatenate((X1, X2), axis=0)
        Y = np.ones(n1 + n2)

        pair = self._convert_pair((label1, label2))
        if pair[0] == label1:
            Y[:n1] = -1
        else:
            Y[n1:] = -1

        # class weights
        weights = self._get_weights(n1, n2)

        # instantiate and fit
        clf = sksvm.SVC(class_weight=weights, **self._clf_kwargs)
        clf.fit(X, Y)

        # add to models
        self._models[pair] = clf

    def _del_clf(self, pair):
        """Delete a binary classifier.

        Parameters
        ----------
        pair : list, tuple
            Label pair.

        """

        pair = self._convert_pair(pair)
        m = self._models.pop(pair)
        del m

    def _convert_pair(self, pair):
        """Sort and convert a label pair to the internal representation format.

        Parameters
        ----------
        pair : list, tuple
            Input label pair.

        Returns
        -------
        pair : tuple
            Sorted label pair.

        """

        pair = tuple(sorted(pair))

        return pair

    def _predict(self, pair, X):
        """Get a classifier prediction of the input data, given the label pair.

        Parameters
        ----------
        pair : list, tuple
            Label pair.
        X : array
            Input data to classify.

        Returns
        -------
        prediction : array
            Prediction for each sample in the input data.

        """

        # convert pair
        pair = self._convert_pair(pair)

        # classify
        aux = self._models[pair].predict(X)

        prediction = []
        for item in aux:
            if item < 0:
                prediction.append(pair[0])
            elif item > 0:
                prediction.append(pair[1])
            else:
                prediction.append('')

        prediction = np.array(prediction)

        return prediction

    def _authenticate(self, data, label, threshold):
        """Authenticate a set of feature vectors, allegedly belonging to the
        given subject.

        Parameters
        ----------
        data : array
            Input test data.
        label : str
            Internal classifier subject label.
        threshold : int, float
            Authentication threshold.

        Returns
        -------
        decision : array
            Authentication decision for each input sample.

        """

        # unpack prepared data
        aux = data['predictions']
        ns = aux.shape[1]
        pairs = data['pairs']

        # normalization
        if self._nbSubjects > 1:
            norm = float(self._nbSubjects - 1)
        else:
            norm = 1.0

        # select pairs
        sel = np.nonzero([label in p for p in pairs])[0]
        aux = aux[sel, :]

        decision = []
        for i in range(ns):
            # determine majority
            predMax, count = majority_rule(aux[:, i], random=True)
            rate = float(count) / norm

            if predMax == '':
                decision.append(False)
            else:
                # compare with threshold
                if rate > threshold:
                    decision.append(predMax == label)
                else:
                    decision.append(False)

        decision = np.array(decision)

        return decision

    def _get_thresholds(self):
        """Generate an array of reasonable thresholds.

        The thresholds correspond to the relative number of binary classifiers
        that agree on a class.

        Returns
        -------
        ths : array
            Generated thresholds.

        """

        ths = np.linspace(self.min_thr, 1.0, 100)

        return ths

    def _identify(self, data, threshold=None):
        """Identify a set of feature vectors.

        Parameters
        ----------
        data : array
            Input test data.
        threshold : int, float
            Identification threshold.

        Returns
        -------
        labels : list
            Identity (internal label) of each input sample.

        """

        if threshold is None:
            thrFcn = lambda label: self.get_id_thr(label, ready=True)
        else:
            thrFcn = lambda label: threshold

        # unpack prepared data
        aux = data['predictions']
        ns = aux.shape[1]

        # normalization
        if self._nbSubjects > 1:
            norm = float(self._nbSubjects - 1)
        else:
            norm = 1.0

        labels = []
        for i in range(ns):
            # determine majority
            predMax, count = majority_rule(aux[:, i], random=True)
            rate = float(count) / norm

            if predMax == '':
                labels.append('')
            else:
                # compare with threshold
                if rate > thrFcn(predMax):
                    # accept
                    labels.append(predMax)
                else:
                    # reject
                    labels.append('')

        return labels

    def _prepare(self, data, targets=None):
        """Prepare data to be processed.

        Computes the predictions for each of the targeted classifier pairs.

        Parameters
        ----------
        data : array
            Data to process.
        targets : list, str, optional
            Target subject labels.

        Returns
        -------
        out : dict
            Processed data containing an array with the predictions of each
            input sample (`predictions`) and a list with the target label
            pairs (`pairs`).

        """

        # target class labels
        if self._nbSubjects == 1:
            pairs = list(self._models)
        else:
            if targets is None:
                pairs = list(self._models)
            elif isinstance(targets, six.string_types):
                labels = list(
                    set(self._subject2label.values()) - set([targets]))
                pairs = [[targets, lbl] for lbl in labels]
            else:
                pairs = []
                for t in targets:
                    labels = list(set(self._subject2label.values()) - set([t]))
                    pairs.extend([t, lbl] for lbl in labels)

        # predict
        predictions = np.array([self._predict(p, data) for p in pairs])

        out = {'predictions': predictions, 'pairs': pairs}

        return out

    def _train(self, enroll=None, dismiss=None):
        """Train the classifier.

        Parameters
        ----------
        enroll : list, optional
            Labels of new or updated subjects.
        dismiss : list, optional
            Labels of deleted subjects.

        """

        if enroll is None:
            enroll = []
        if dismiss is None:
            dismiss = []

        # process dismiss
        src_pairs = list(self._models)
        pairs = []
        for t in dismiss:
            pairs.extend([p for p in src_pairs if t in p])

        for p in pairs:
            self._del_clf(p)

        # process enroll
        existing = list(set(self._subject2label.values()) - set(enroll))
        for i, t1 in enumerate(enroll):
            X1 = self.io_load(t1)
            n1 = len(X1)

            # existing subjects
            for t2 in existing:
                X2 = self.io_load(t2)
                n2 = len(X2)
                self._get_kernel_clf(X1, X2, n1, n2, t1, t2)

            # new subjects
            for t2 in enroll[i + 1:]:
                X2 = self.io_load(t2)
                n2 = len(X2)
                self._get_kernel_clf(X1, X2, n1, n2, t1, t2)

        # check singles
        if self._nbSubjects == 1:
            label = list(six.itervalues(self._subject2label))[0]
            X = self.io_load(label)
            self._get_single_clf(X, label)
        elif self._nbSubjects > 1:
            aux = [p for p in self._models if '' in p]
            if len(aux) != 0:
                for p in aux:
                    self._del_clf(p)

    def _update(self, old, new):
        """Combine new data with existing templates (for one subject).

        Simply concatenates old data with new data.

        Parameters
        ----------
        old : array
            Existing data.
        new : array
            New data.

        Returns
        -------
        out : array
            Combined data.

        """

        out = np.concatenate([old, new], axis=0)

        return out


def get_auth_rates(TP=None, FP=None, TN=None, FN=None, thresholds=None):
    """Compute authentication rates from the confusion matrix.

    Parameters
    ----------
    TP : array
        True Positive counts for each classifier threshold.
    FP : array
        False Positive counts for each classifier threshold.
    TN : array
        True Negative counts for each classifier threshold.
    FN : array
        False Negative counts for each classifier threshold.
    thresholds : array
        Classifier thresholds.

    Returns
    -------
    Acc : array
        Accuracy at each classifier threshold.
    TAR : array
        True Accept Rate at each classifier threshold.
    FAR : array
        False Accept Rate at each classifier threshold.
    FRR : array
        False Reject Rate at each classifier threshold.
    TRR : array
        True Reject Rate at each classifier threshold.
    EER : array
        Equal Error Rate points, with format (threshold, rate).
    Err : array
        Error rate at each classifier threshold.
    PPV : array
        Positive Predictive Value at each classifier threshold.
    FDR : array
        False Discovery Rate at each classifier threshold.
    NPV : array
        Negative Predictive Value at each classifier threshold.
    FOR : array
        False Omission Rate at each classifier threshold.
    MCC : array
        Matthrews Correlation Coefficient at each classifier threshold.

    """

    # check inputs
    if TP is None:
        raise TypeError("Please specify the input TP counts.")
    if FP is None:
        raise TypeError("Please specify the input FP counts.")
    if TN is None:
        raise TypeError("Please specify the input TN counts.")
    if FN is None:
        raise TypeError("Please specify the input FN counts.")
    if thresholds is None:
        raise TypeError("Please specify the input classifier thresholds.")

    # ensure numpy
    TP = np.array(TP)
    FP = np.array(FP)
    TN = np.array(TN)
    FN = np.array(FN)
    thresholds = np.array(thresholds)

    # helper variables
    A = TP + FP
    B = TP + FN
    C = TN + FP
    D = TN + FN
    E = A * B * C * D
    F = A + D

    # avoid divisions by zero
    A[A == 0] = 1.
    B[B == 0] = 1.
    C[C == 0] = 1.
    D[D == 0] = 1.
    E[E == 0] = 1.
    F[F == 0] = 1.

    # rates
    Acc = (TP + TN) / F # accuracy
    Err = (FP + FN) / F # error rate

    TAR = TP / B # true accept rate /true positive rate
    FRR = FN / B # false rejection rate / false negative rate

    TRR = TN / C # true rejection rate / true negative rate
    FAR = FP / C # false accept rate / false positive rate

    PPV = TP / A # positive predictive value
    FDR = FP / A # false discovery rate

    NPV = TN / D # negative predictive value
    FOR = FN / D # false omission rate

    MCC = (TP*TN - FP*FN) / np.sqrt(E) # matthews correlation coefficient 

    # determine EER
    roots, values = tools.find_intersection(thresholds, FAR, thresholds, FRR)
    EER = np.vstack((roots, values)).T

    # output
    args = (Acc, TAR, FAR, FRR, TRR, EER, Err, PPV, FDR, NPV, FOR, MCC)
    names = ('Acc', 'TAR', 'FAR', 'FRR', 'TRR', 'EER', 'Err', 'PPV', 'FDR',
             'NPV', 'FOR', 'MCC')

    return utils.ReturnTuple(args, names)


def get_id_rates(H=None, M=None, R=None, N=None, thresholds=None):
    """Compute identification rates from the confusion matrix.

    Parameters
    ----------
    H : array
        Hit counts for each classifier threshold.
    M : array
        Miss counts for each classifier threshold.
    R : array
        Reject counts for each classifier threshold.
    N : int
        Number of test samples.
    thresholds : array
        Classifier thresholds.

    Returns
    -------
    Acc : array
        Accuracy at each classifier threshold.
    Err : array
        Error rate at each classifier threshold.
    MR : array
        Miss Rate at each classifier threshold.
    RR : array
        Reject Rate at each classifier threshold.
    EID : array
        Error of Identification points, with format (threshold, rate).
    EER : array
        Equal Error Rate points, with format (threshold, rate).

    """

    # check inputs
    if H is None:
        raise TypeError("Please specify the input H counts.")
    if M is None:
        raise TypeError("Please specify the input M counts.")
    if R is None:
        raise TypeError("Please specify the input R counts.")
    if N is None:
        raise TypeError("Please specify the total number of test samples.")
    if thresholds is None:
        raise TypeError("Please specify the input classifier thresholds.")

    # ensure numpy
    H = np.array(H)
    M = np.array(M)
    R = np.array(R)
    thresholds = np.array(thresholds)

    Acc = H / N
    Err = 1 - Acc
    MR = M / N
    RR = R / N

    # EER
    roots, values = tools.find_intersection(thresholds, MR, thresholds, RR)
    EER = np.vstack((roots, values)).T

    # EID
    y2 = np.min(Err) * np.ones(len(thresholds), dtype='float')
    roots, values = tools.find_intersection(thresholds, Err, thresholds, y2)
    EID = np.vstack((roots, values)).T

    # output
    args = (Acc, Err, MR, RR, EID, EER)
    names = ('Acc', 'Err', 'MR', 'RR', 'EID', 'EER')

    return utils.ReturnTuple(args, names)


def get_subject_results(results=None,
                        subject=None,
                        thresholds=None,
                        subjects=None,
                        subject_dict=None,
                        subject_idx=None):
    """Compute authentication and identification performance metrics for a
    given subject.

    Parameters
    ----------
    results : dict
        Classification results.
    subject : hashable
        True subject label.
    thresholds : array
        Classifier thresholds.
    subjects : list
        Target subject classes.
    subject_dict : bidict
        Subject-label conversion dictionary.
    subject_idx : list
        Subject index.

    Returns
    -------
    assessment : dict
        Authentication and identification results.

    """

    # check inputs
    if results is None:
        raise TypeError("Please specify the input classification results.")
    if subject is None:
        raise TypeError("Please specify the input subject class.")
    if thresholds is None:
        raise TypeError("Please specify the input classifier thresholds.")
    if subjects is None:
        raise TypeError("Please specify the target subject classes.")
    if subject_dict is None:
        raise TypeError("Please specify the subject-label dictionary.")
    if subject_idx is None:
        raise TypeError("Plase specify subject index.")

    nth = len(thresholds)
    auth_res = results['authentication']
    id_res = results['identification']
    ns = auth_res.shape[2]

    # sanity checks
    if auth_res.shape[0] != id_res.shape[0]:
        raise ValueError("Authentication and identification number of \
                          thresholds do not match.")
    if auth_res.shape[0] != nth:
        raise ValueError("Number of thresholds in vector does not match \
                          biometric results.")
    if auth_res.shape[2] != id_res.shape[1]:
        raise ValueError("Authentication and identification number of tests \
                          do not match.")

    label = subject_dict[subject]

    # authentication vars
    TP = np.zeros(nth, dtype='float')
    FP = np.zeros(nth, dtype='float')
    TN = np.zeros(nth, dtype='float')
    FN = np.zeros(nth, dtype='float')

    # identification vars
    H = np.zeros(nth, dtype='float')
    M = np.zeros(nth, dtype='float')
    R = np.zeros(nth, dtype='float')
    CM = []

    for i in range(nth):  # for each threshold
        # authentication
        for k, lbl in enumerate(subject_idx):  # for each subject
            subject_tst = subjects[k]

            d = auth_res[i, lbl, :]
            if subject == subject_tst:
                # true positives
                aux = np.sum(d)
                TP[i] += aux
                # false negatives
                FN[i] += (ns - aux)
            else:
                # false positives
                aux = np.sum(d)
                FP[i] += aux
                # true negatives
                TN[i] += (ns - aux)

        # identification
        res = id_res[i, :]
        hits = res == label
        nhits = np.sum(hits)
        rejects = res == ''
        nrejects = np.sum(rejects)
        misses = np.logical_not(np.logical_or(hits, rejects))
        nmisses = ns - (nhits + nrejects)
        missCounts = {
            subject_dict.inv[ms]: np.sum(res == ms)
            for ms in np.unique(res[misses])
        }

        # appends
        H[i] = nhits
        M[i] = nmisses
        R[i] = nrejects
        CM.append(missCounts)

    # compute rates
    auth_rates = get_auth_rates(TP, FP, TN, FN, thresholds).as_dict()
    id_rates = get_id_rates(H, M, R, ns, thresholds).as_dict()

    output = {
        'authentication': {
            'confusionMatrix': {'TP': TP, 'FP': FP, 'TN': TN, 'FN': FN},
            'rates': auth_rates,
        },
        'identification': {
            'confusionMatrix': {'H': H, 'M': M, 'R': R, 'CM': CM},
            'rates': id_rates,
        },
    }

    return utils.ReturnTuple((output,), ('assessment',))


def assess_classification(results=None, thresholds=None):
    """Assess the performance of a biometric classification test.

    Parameters
    ----------
    results : dict
        Classification results.
    thresholds : array
        Classifier thresholds.

    Returns
    -------
    assessment : dict
        Classification assessment.

    """

    # check inputs
    if results is None:
        raise TypeError("Please specify the input classification results.")
    if thresholds is None:
        raise TypeError("Please specify the input classifier thresholds.")

    # test subjects
    subjectDict = results['subjectDict']
    subParent = results['subjectList']
    subIdx = [subParent.index(item) for item in subParent]
    subIdx.sort()
    subjects = [subParent[item] for item in subIdx]

    # output object
    output = {
        'global': {
            'authentication': {
                'confusionMatrix': {'TP': 0., 'TN': 0., 'FP': 0., 'FN': 0.},
            },
            'identification': {
                'confusionMatrix': {'H': 0., 'M': 0., 'R': 0.},
            },
        },
        'subject': {},
        'thresholds': thresholds,
    }

    nth = len(thresholds)
    C = np.zeros((nth, len(subjects)), dtype='float')

    # update variables
    auth = output['global']['authentication']['confusionMatrix']
    authM = ['TP', 'TN', 'FP', 'FN']
    iden = output['global']['identification']['confusionMatrix']
    idenM = ['H', 'M', 'R']

    for test_user in subjects:
        aux, = get_subject_results(results[test_user], test_user, thresholds,
                                   subjects, subjectDict, subIdx)

        # copy to subject
        output['subject'][test_user] = aux

        # authentication
        for m in authM:
            auth[m] += aux['authentication']['confusionMatrix'][m]

        # identification
        for m in idenM:
            iden[m] += aux['identification']['confusionMatrix'][m]

        # subject misses
        for i, item in enumerate(aux['identification']['confusionMatrix']['CM']):
            for k, sub in enumerate(subjects):
                try:
                    C[i, k] += item[sub]
                except KeyError:
                    pass

    # normalize subject misses
    sC = C.sum(axis=1).reshape((nth, 1))
    # avoid division by zero
    sC[sC <= 0] = 1.
    CR = C / sC

    # update subjects
    for k, sub in enumerate(subjects):
        output['subject'][sub]['identification']['confusionMatrix']['C'] = C[:,
                                                                             k]
        output['subject'][sub]['identification']['rates']['CR'] = CR[:, k]

    # compute global rates
    aux = get_auth_rates(auth['TP'], auth['FP'], auth['TN'], auth['FN'],
                         thresholds)
    output['global']['authentication']['rates'] = aux.as_dict()

    # identification
    Ns = iden['H'] + iden['M'] + iden['R']
    aux = get_id_rates(iden['H'], iden['M'], iden['R'], Ns, thresholds)
    output['global']['identification']['rates'] = aux.as_dict()

    return utils.ReturnTuple((output,), ('assessment',))


def assess_runs(results=None, subjects=None):
    """Assess the performance of multiple biometric classification runs.

    Parameters
    ----------
    results : list
        Classification assessment for each run.
    subjects : list
        Common target subject classes.

    Returns
    -------
    assessment : dict
        Global classification assessment.

    """

    # check inputs
    if results is None:
        raise TypeError("Please specify the input classification results.")
    if subjects is None:
        raise TypeError("Please specify the common subject classes.")

    nb = len(results)
    if nb == 0:
        raise ValueError("Please provide at least one classification run.")
    elif nb == 1:
        return utils.ReturnTuple((results[0],), ('assessment',))

    # output
    output = {
        'global': {
            'authentication': {
                'confusionMatrix': {'TP': 0., 'TN': 0., 'FP': 0., 'FN': 0.},
            },
            'identification': {
                'confusionMatrix': {'H': 0., 'M': 0., 'R': 0.},
            },
        },
        'subject': {},
        'thresholds': None,
    }

    thresholds = output['thresholds'] = results[0]['thresholds']

    # global helpers
    auth = output['global']['authentication']['confusionMatrix']
    iden = output['global']['identification']['confusionMatrix']
    authM = ['TP', 'TN', 'FP', 'FN']
    idenM1 = ['H', 'M', 'R', 'C']
    idenM2 = ['H', 'M', 'R']

    for sub in subjects:
        # create subject confusion matrix, rates
        output['subject'][sub] = {
            'authentication': {
                'confusionMatrix': {'TP': 0., 'TN': 0., 'FP': 0., 'FN': 0.},
                'rates': {},
            },
            'identification': {
                'confusionMatrix': {'H': 0., 'M': 0., 'R': 0., 'C': 0.},
                'rates': {},
            },
        }

        # subject helpers
        authS = output['subject'][sub]['authentication']['confusionMatrix']
        idenS = output['subject'][sub]['identification']['confusionMatrix']

        # update confusions
        for run in results:
            # authentication
            auth_run = run['subject'][sub]['authentication']['confusionMatrix']
            for m in authM:
                auth[m] += auth_run[m]
                authS[m] += auth_run[m]

            # identification
            iden_run = run['subject'][sub]['identification']['confusionMatrix']
            for m in idenM1:
                idenS[m] += iden_run[m]
            for m in idenM2:
                iden[m] += iden_run[m]

        # compute subject mean
        # authentication
        for m in authM:
            authS[m] /= float(nb)

        # identification
        for m in idenM1:
            idenS[m] /= float(nb)

        # compute subject rates
        aux = get_auth_rates(authS['TP'], authS['FP'], authS['TN'],
                             authS['FN'], thresholds)
        output['subject'][sub]['authentication']['rates'] = aux.as_dict()

        Ns = idenS['H'] + idenS['M'] + idenS['R']
        aux = get_id_rates(idenS['H'], idenS['M'], idenS['R'], Ns, thresholds)
        output['subject'][sub]['identification']['rates'] = aux.as_dict()
        M = np.array(idenS['M'], copy=True)
        M[M <= 0] = 1.
        output['subject'][sub]['identification']['rates']['CR'] = idenS['C'] / M

    # compute global mean
    # authentication
    for m in authM:
        auth[m] /= float(nb)

    # identification
    for m in idenM2:
        iden[m] /= float(nb)

    # compute rates
    aux = get_auth_rates(auth['TP'], auth['FP'], auth['TN'], auth['FN'],
                         thresholds)
    output['global']['authentication']['rates'] = aux.as_dict()

    Ns = iden['H'] + iden['M'] + iden['R']
    aux = get_id_rates(iden['H'], iden['M'], iden['R'], Ns, thresholds)
    output['global']['identification']['rates'] = aux.as_dict()

    return utils.ReturnTuple((output,), ('assessment',))


def combination(results=None, weights=None):
    """Combine results from multiple classifiers.

    Parameters
    ----------
    results : dict
        Results for each classifier.
    weights : dict, optional
        Weight for each classifier.

    Returns
    -------
    decision : object
        Consensus decision.
    confidence : float
        Confidence estimate of the decision.
    counts : array
        Weight for each possible decision outcome.
    classes : array
        List of possible decision outcomes.

    """

    # check inputs
    if results is None:
        raise TypeError("Please specify the input classification results.")
    if weights is None:
        weights = {}

    # compile results to find all classes
    vec = list(six.itervalues(results))
    if len(vec) == 0:
        raise CombinationError("No keys found.")

    unq = np.unique(np.concatenate(vec))

    nb = len(unq)
    if nb == 0:
        # empty array
        raise CombinationError("No values found.")
    elif nb == 1:
        # unanimous result
        decision = unq[0]
        confidence = 1.
        counts = [1.]
    else:
        # multi-class
        counts = np.zeros(nb, dtype='float')

        for n in results:
            # ensure array
            res = np.array(results[n])
            ns = float(len(res))

            # get count for each unique class
            for i in range(nb):
                aux = float(np.sum(res == unq[i]))
                w = weights.get(n, 1.)
                counts[i] += ((aux / ns) * w)

        # most frequent class
        predMax = counts.argmax()
        counts /= counts.sum()

        decision = unq[predMax]
        confidence = counts[predMax]

    # output
    args = (decision, confidence, counts, unq)
    names = ('decision', 'confidence', 'counts', 'classes')

    return utils.ReturnTuple(args, names)


def majority_rule(labels=None, random=True):
    """Determine the most frequent class label.

    Parameters
    ----------
    labels : array, list
        List of clas labels.
    random : bool, optional
        If True, will choose randomly in case of tied classes, otherwise the
        first element is chosen.

    Returns
    -------
        decision : object
            Consensus decision.
        count : int
            Number of elements of the consensus decision.

    """

    # check inputs
    if labels is None:
        raise TypeError("Please specify the input list of class labels.")

    if len(labels) == 0:
        raise CombinationError("Empty list of class labels.")

    # count unique occurrences
    unq, counts = np.unique(labels, return_counts=True)

    # most frequent class
    predMax = counts.argmax()

    if random:
        # check for repeats
        ind = np.nonzero(counts == counts[predMax])[0]
        length = len(ind)

        if length > 1:
            predMax = ind[np.random.randint(0, length)]

    decision = unq[predMax]
    cnt = counts[predMax]

    out = utils.ReturnTuple((decision, cnt), ('decision', 'count'))

    return out


def cross_validation(labels,
                     n_iter=10,
                     test_size=0.1,
                     train_size=None,
                     random_state=None):
    """Return a Cross Validation (CV) iterator.

    Wraps the StratifiedShuffleSplit iterator from sklearn.model_selection.
    This iterator returns stratified randomized folds, which preserve the
    percentage of samples for each class.

    Parameters
    ----------
    labels : list, array
        List of class labels for each data sample.
    n_iter : int, optional
        Number of splitting iterations.
    test_size : float, int, optional
        If float, represents the proportion of the dataset to include in the
        test split; if int, represents the absolute number of test samples.
    train_size : float, int, optional
        If float, represents the proportion of the dataset to include in the
        train split; if int, represents the absolute number of train samples.
    random_state : int, RandomState, optional
        The seed of the pseudo random number generator to use when shuffling
        the data.

    Returns
    -------
    cv : CV iterator
        Cross Validation iterator.

    """

    cv = skcv.StratifiedShuffleSplit(
        n_splits=n_iter,
        test_size=test_size,
        train_size=train_size,
        random_state=random_state,
    ).split(np.zeros(len(labels)), labels)

    return utils.ReturnTuple((cv,), ('cv',))
